from sys import modules as __modules

from .exceptions import UwsgiconfException as __DummyException
from .settings import FORCE_STUB as __FORCE_STUB

if False:  # pragma: nocover
    from .uwsgi_stub import *  # noqa  # Give IDEs a chance to load stub symbols.


try:  # pragma: nocover
    if __FORCE_STUB:
        raise __DummyException('uWSGI stub instruction is found in env.')

    import uwsgi

    uwsgi.is_stub = False
    """Indicates whether stub is used instead of real `uwsgi` module."""

    # The following allows proper dynamic attributes (e.g. ``env``) addressing.
    __modules[__name__] = uwsgi

except (ImportError, __DummyException):

    from . import uwsgi_stub
    __modules[__name__] = uwsgi_stub



from threading import local
from typing import Callable, Type, Tuple, Union, Optional, Dict, List

from .request import _Request
from .. import uwsgi as _uwsgi
from ..utils import decode_deep, decode


class _PostForkHooks:

    funcs = []

    @classmethod
    def add(cls) -> Callable:
        """Decorator. Registers a function to be called after process fork()."""

        def add_(func: Callable):
            cls.funcs.append(func)
            return func

        return add_

    @classmethod
    def run(cls):
        """Runs registered hooks."""
        for func in cls.funcs:
            func()


_uwsgi.post_fork_hook = _PostForkHooks.run


class _Platform:

    request: Type[_Request] = _Request
    """Current request information. 
    
    .. note:: Object is attached runtime.
    
    """

    postfork_hooks = _PostForkHooks
    """uWSGI is a preforking server, so you might need 
    to execute a fixup tasks (hooks) after each fork(). 
    Each hook will be executed in sequence on each process (worker/mule).
    
    .. note:: The fork() happen before app loading, so there's no hooks for dynamic apps.
        But one can still move postfork hooks in a .py file and 
        import it on server startup with `python.import_module()`.

    """

    workers_count: int = _uwsgi.numproc
    """Number of workers (processes) currently running."""

    cores_count: int = _uwsgi.cores
    """Detected number of processor cores."""

    buffer_size: int = _uwsgi.buffer_size
    """The current configured buffer size in bytes."""

    threads_enabled: bool = _uwsgi.has_threads
    """Flag indicating whether thread support is enabled."""

    started_on: int = _uwsgi.started_on
    """uWSGI's startup Unix timestamp."""

    apps_map: Optional[dict] = _uwsgi.applications
    """Applications dictionary mapping mountpoints to application callables."""

    @property
    def hostname(self) -> str:
        """Current host name."""
        return decode(_uwsgi.hostname)

    @property
    def config(self) -> Dict[str, Union[str, List[str]]]:
        """The current configuration options, including any custom placeholders."""
        return decode_deep(_uwsgi.opt)

    @property
    def config_variables(self) -> Dict[str, str]:
        """Current mapping of configuration file "magic" variables."""
        return decode_deep(_uwsgi.magic_table)

    @property
    def worker_id(self) -> int:
        """Returns current worker ID. 0 if not a worker (e.g. mule)."""
        return _uwsgi.worker_id()

    @property
    def workers_info(self) -> Tuple[dict, ...]:
        """Gets statistics for all the workers for the current server.

        Returns tuple of dicts.

        """
        return tuple(decode_deep(item) for item in _uwsgi.workers())

    @property
    def ready_for_requests(self) -> bool:
        """Returns flag indicating whether we are ready to handle requests."""
        return _uwsgi.ready()

    @property
    def master_pid(self) -> int:
        """Return the process identifier (PID) of the uWSGI master process."""
        return _uwsgi.masterpid()

    @property
    def memory(self) -> Tuple[int, int]:
        """Returns memory usage tuple of ints: (rss, vsz)."""
        return _uwsgi.mem()

    @property
    def clock(self) -> int:
        """Returns uWSGI clock microseconds."""
        return _uwsgi.micros()

    def get_listen_queue(self, socket_num: int = 0) -> int:
        """Returns listen queue (backlog size) of the given socket.

        :param socket_num: Socket number.

        :raises ValueError: If socket is not found

        """
        return _uwsgi.listen_queue(socket_num)

    def get_version(self, *, as_tuple: bool = False) -> Union[str, Tuple[int, int, int, int, str]]:
        """Returns uWSGI version string or tuple.

        :param as_tuple:

        """
        if as_tuple:
            return decode_deep(_uwsgi.version_info)

        return decode(_uwsgi.version)


__THREAD_LOCAL = local()


def __get_platform() -> _Platform:
    platform = getattr(__THREAD_LOCAL, 'uwsgi_platform', None)

    if platform is None:
        platform = _Platform()
        platform.request = _Request()

        setattr(__THREAD_LOCAL, 'uwsgi_platform', platform)

    return platform


uwsgi = __get_platform()
"""This is a _Platform instance."""



from pathlib import Path
from typing import Union

from ..base import ParametrizedValue
from ..utils import KeyValue, filter_locals


class RouteAction(ParametrizedValue):

    pass


class ActionFlush(RouteAction):
    """Send the current contents of the transformation buffer
    to the client (without clearing the buffer).

    * http://uwsgi.readthedocs.io/en/latest/Transformations.html#flushing-magic

    """
    name = 'flush'

    def __init__(self):
        super().__init__()


class ActionGzip(RouteAction):
    """Encodes the response buffer to gzip."""

    name = 'gzip'
    plugin = 'transformation_gzip'

    def __init__(self):
        super().__init__()


class ActionToFile(RouteAction):
    """Used for caching a response buffer into a static file."""

    name = 'tofile'
    plugin = 'transformation_tofile'

    def __init__(self, filename, *, mode=None):
        arg = KeyValue(locals())
        super().__init__(arg)


class ActionUpper(RouteAction):
    """Transforms each character in uppercase.

    Mainly as an example of transformation plugin.

    """
    name = 'toupper'
    plugin = 'transformation_toupper'

    def __init__(self):
        super().__init__()


class ActionChunked(RouteAction):
    """Encodes the output in HTTP chunked."""

    name = 'chunked'

    def __init__(self):
        super().__init__()


class ActionTemplate(RouteAction):
    """Allows using a template file to expose everything
    from internal routing system into it.

    * http://uwsgi.readthedocs.io/en/latest/Changelog-1.9.19.html#the-template-transformation

    """
    name = 'template'

    def __init__(self):
        super().__init__()


class ActionFixContentLen(RouteAction):
    """Fixes Content-length header."""

    name = 'fixcl'

    def __init__(self, *, add_header=False):
        """
        :param bool add_header: Force header add instead of plain fix of existing header.
        """
        if add_header:
            self.name = 'forcexcl'

        super().__init__()


class ActionDoContinue(RouteAction):
    """Stop scanning the internal routing table
    and continue to the selected request handler.

    """
    name = 'continue'

    def __init__(self):
        super().__init__()


class ActionDoBreak(RouteAction):
    """Stop scanning the internal routing table and close the request."""

    name = 'break'

    def __init__(self, code, *, return_body=False):
        """
        :param int code: HTTP code

        :param return_body: Uses uWSGI's built-in status code and returns
            both status code and message body.

        """
        if return_body:
            self.name = 'return'

        super().__init__(code)


class ActionLog(RouteAction):
    """Print the specified message in the logs or
    do not log a request is ``message`` is ``None``.

    """
    name = 'log'

    def __init__(self, message):
        """

        :param str|None message: Message to add into log.
            If ``None`` logging will be disabled for this request.

        """
        if message is None:
            self.name = 'donotlog'

        super().__init__(message)


class ActionOffloadOff(RouteAction):
    """Do not use offloading."""

    name = 'donotoffload'

    def __init__(self):
        super().__init__()


class ActionAddVarLog(RouteAction):
    """Add the specified logvar."""

    name = 'logvar'

    def __init__(self, name, val):
        """
        :param str name: Variable name.

        :param val: Variable value.
        """
        super().__init__(name, val)


class ActionDoGoto(RouteAction):
    """Make a forward jump to the specified label or rule position."""

    name = 'goto'

    def __init__(self, where):
        """
        :param str|int where: Rule number of label to go to.
        """
        super().__init__(where)


class ActionAddVarCgi(RouteAction):
    """Add the specified CGI (environment) variable to the request."""

    name = 'addvar'

    def __init__(self, name, val):
        """
        :param str name: Variable name.

        :param val: Variable value.
        """
        super().__init__(name, val)


class ActionHeaderAdd(RouteAction):
    """Add the specified HTTP header to the response."""

    name = 'addheader'

    def __init__(self, name, val):
        """
        :param str name: Header name.

        :param val: Header value.
        """
        name += ':'

        super().__init__(name, val)


class ActionHeaderRemove(RouteAction):
    """Remove the specified HTTP header from the response."""

    name = 'delheader'

    def __init__(self, name):
        """
        :param str name: Header name.
        """
        super().__init__(name)


class ActionHeadersOff(RouteAction):
    """Disable headers."""

    name = 'disableheaders'

    def __init__(self):
        super().__init__()


class ActionHeadersReset(RouteAction):
    """Clear the response headers, setting a new HTTP status code,
    useful for resetting a response.

    """
    name = 'clearheaders'

    def __init__(self, code):
        """
        :param int code: HTTP code.
        """
        super().__init__(code)


class ActionSignal(RouteAction):
    """Raise the specified uwsgi signal."""

    name = 'signal'

    def __init__(self, num):
        """
        :param int num: Signal number.
        """
        super().__init__(num)


class ActionSend(RouteAction):
    """Extremely advanced (and dangerous) function allowing you
    to add raw data to the response.

    """
    name = 'send'

    def __init__(self, data, *, crnl: bool = False):
        """
        :param data: Data to add to response.
        :param crnl: Add carriage return and new line.

        """
        if crnl:
            self.name = 'send-crnl'
        super().__init__(data)


class ActionRedirect(RouteAction):
    """Return a HTTP 301/302 Redirect to the specified URL."""

    name = 'redirect-302'
    plugin = 'router_redirect'

    def __init__(self, url, *, permanent=False):
        """
        :param str url: URL to redirect to.
        :param bool permanent: If ``True`` use 301, otherwise 302.
        """
        if permanent:
            self.name = 'redirect-301'

        super().__init__(url)


class ActionRewrite(RouteAction):
    """A rewriting engine inspired by Apache mod_rewrite.

    Rebuild PATH_INFO and QUERY_STRING according to the specified rules
    before the request is dispatched to the request handler.

    """
    name = 'rewrite'
    plugin = 'router_rewrite'

    def __init__(self, rule, *, do_continue=False):
        """
        :param str rule: A rewrite rule.

        :param bool do_continue: Stop request processing
            and continue to the selected request handler.

        """
        if do_continue:
            self.name = 'rewrite-last'

        super().__init__(rule)


class ActionRouteUwsgi(RouteAction):
    """Rewrite the modifier1, modifier2 and optionally UWSGI_APPID values of a request
    or route the request to an external uwsgi server.

    """
    name = 'uwsgi'
    plugin = 'router_uwsgi'
    args_joiner = ','

    def __init__(self, external_address='', *, modifier='', app=''):
        """
        :param str external_address: External uWSGI server address (host:port).
        :param Modifier modifier: Set request modifier.
        :param str app: Set ``UWSGI_APPID``.

        """
        super().__init__(external_address, modifier, modifier.submod, app)


class ActionRouteExternal(RouteAction):
    """Route the request to an external HTTP server."""

    name = 'http'
    plugin = 'router_http'
    args_joiner = ','

    def __init__(self, address, *, host_header=None):
        """
        :param str address: External HTTP address (host:port)

        :param str host_header: HOST header value.

        """
        super().__init__(address, host_header)


class ActionAlarm(RouteAction):
    """Triggers an alarm.

    * http://uwsgi.readthedocs.io/en/latest/Changelog-1.9.6.html#the-alarm-routing-action

    """
    name = 'alarm'

    def __init__(self, name, message):
        """
        :param str name: Alarm name

        :param str message: Message to pass into alarm.
        """
        super().__init__(name, message)


class ActionServeStatic(RouteAction):
    """Serve a static file from the specified physical path."""

    name = 'static'
    plugin = 'router_static'

    def __init__(self, fpath: Union[str, Path]):
        """
        :param fpath: Static file path.

        """
        super().__init__(str(fpath))


class ActionAuthBasic(RouteAction):
    """Use Basic HTTP Auth."""

    name = 'basicauth'
    plugin = 'router_basicauth'
    args_joiner = ','

    def __init__(self, realm, *, user=None, password=None, do_next=False):
        """
        :param str realm:

        :param str user:

        :param str password: Password or htpasswd-like file.

        :param bool do_next: Allow next rule.
        """
        if do_next:
            self.name = 'basicauth-next'

        user_password = []

        if user:
            user += ':'
            user_password.append(user)

        if password:
            user_password.append(password)

        super().__init__(realm, ''.join(user_password) if user_password else None)


class AuthLdap(RouteAction):
    """Use Basic HTTP Auth."""

    name = 'ldapauth'
    plugin = 'ldap'
    args_joiner = ','

    def __init__(
            self, realm, address, *, base_dn=None, bind_dn=None, bind_password=None,
            filter=None, login_attr=None, log_level=None,
            do_next=False):
        """

        :param str realm:

        :param str address: LDAP server URI

        :param str base_dn: Base DN used when searching for users.

        :param str bind_dn: DN used for binding.
            Required if the LDAP server does not allow anonymous searches.

        :param str bind_password: Password for the ``bind_dn`` user.

        :param str filter: Filter used when searching for users. Default: ``(objectClass=*)``

        :param str login_attr: LDAP attribute that holds user login. Default: ``uid``.

        :param str log_level: Log level.

            Supported values:
                * 0 - don't log any binds
                * 1 - log authentication errors,
                * 2 - log both successful and failed binds

        :param bool do_next: Allow next rule.
        """
        arg = KeyValue(
            filter_locals(locals(), drop=['realm', 'do_next']),
            aliases={
                'address': 'url',
                'base_dn': 'basedn',
                'bind_dn': 'binddn',
                'bind_password': 'bindpw',
                'login_attr': 'attr',
                'log_level': 'loglevel',
            },
            items_separator=';'
        )

        if do_next:
            self.name = 'ldapauth-next'

        super().__init__(realm, arg)


class ActionSetHarakiri(RouteAction):
    """Set harakiri timeout for the current request."""

    name = 'harakiri'

    def __init__(self, timeout):
        """
        :param int timeout:
        """
        super().__init__(timeout)


class ActionDirChange(RouteAction):
    """Changes a directory."""

    name = 'chdir'

    def __init__(self, dir):
        """
        :param str dir: Directory to change into.
        """
        super().__init__(dir)


class ActionSetVarUwsgiAppid(RouteAction):
    """Set UWSGI_APPID.

    Bypass ``SCRIPT_NAME`` and ``VirtualHosting`` to let the user choose
    the mountpoint without limitations (or headaches).

    The concept is very generic: ``UWSGI_APPID`` is the identifier of an application.
    If it is not found in the internal list of apps, it will be loaded.

    """
    name = 'setapp'

    def __init__(self, app):
        """
        :param str app: Application ID.
        """
        super().__init__(app)


class ActionSetVarRemoteUser(RouteAction):
    """Set REMOTE_USER"""

    name = 'setuser'

    def __init__(self, user):
        """
        :param str user: Username.
        """
        super().__init__(user)


class ActionSetVarUwsgiHome(RouteAction):
    """Set UWSGI_HOME"""

    name = 'sethome'

    def __init__(self, dir):
        """
        :param str dir: Directory to make a new home.
        """
        super().__init__(dir)


class ActionSetVarUwsgiScheme(RouteAction):
    """Set UWSGI_SCHEME.

    Set the URL scheme when it cannot be reliably determined.
    This may be used to force HTTPS (with the value ``https``), for instance.

    * http://uwsgi.readthedocs.io/en/latest/Changelog-1.9.6.html#configuring-dynamic-apps-with-internal-routing

    """
    name = 'setscheme'

    def __init__(self, value):
        """
        :param str value:
        """
        super().__init__(value)


class ActionSetVarScriptName(RouteAction):
    """Set SCRIPT_NAME"""

    name = 'setscriptname'

    def __init__(self, name):
        """
        :param str name: Script name
        """
        super().__init__(name)


class ActionSetVarRequestMethod(RouteAction):
    """Set REQUEST_METHOD"""

    name = 'setmethod'

    def __init__(self, name):
        """
        :param str name: Method name.
        """
        super().__init__(name)


class ActionSetVarRequestUri(RouteAction):
    """Set REQUEST_URI"""

    name = 'seturi'

    def __init__(self, value):
        """
        :param str value: URI
        """
        super().__init__(value)


class ActionSetVarRemoteAddr(RouteAction):
    """Set REMOTE_ADDR"""

    name = 'setremoteaddr'

    def __init__(self, value):
        """
        :param str value: Address.
        """
        super().__init__(value)


class ActionSetVarPathInfo(RouteAction):
    """Set PATH_INFO"""

    name = 'setpathinfo'

    def __init__(self, value):
        """
        :param str value: New info.
        """
        super().__init__(value)


class ActionSetVarDocumentRoot(RouteAction):
    """Set DOCUMENT_ROOT"""

    name = 'setdocroot'

    def __init__(self, value):
        """
        :param str value:
        """
        super().__init__(value)


class ActionSetUwsgiProcessName(RouteAction):
    """Set uWSGI process name."""

    name = 'setprocname'

    def __init__(self, name):
        """
        :param str name: New process name.
        """
        super().__init__(name)


class ActionFixVarPathInfo(RouteAction):
    """Fixes PATH_INFO taking into account script name.

    This action allows you to set SCRIPT_NAME in nginx without bothering
    to rewrite the PATH_INFO (something nginx cannot afford).

    * http://uwsgi.readthedocs.io/en/latest/Changelog-2.0.11.html#fixpathinfo-routing-action

    """
    name = 'fixpathinfo'

    def __init__(self):
        super().__init__()


class ActionSetScriptFile(RouteAction):
    """Set script file.

    * http://uwsgi.readthedocs.io/en/latest/Changelog-1.9.6.html#configuring-dynamic-apps-with-internal-routing

    """
    name = 'setfile'

    def __init__(self, fpath: Union[str, Path]):
        """
        :param str fpath: File path.

        """
        super().__init__(str(fpath))



from uwsgiconf.runtime.platform import uwsgi


def test_platform_stub():

    assert not uwsgi.worker_id
    assert not uwsgi.workers_info
    assert not uwsgi.ready_for_requests
    assert uwsgi.memory == (0, 0)
    assert not uwsgi.clock
    assert not uwsgi.get_listen_queue()
    assert uwsgi.get_version() == '0.0.0'
    assert uwsgi.get_version(as_tuple=True) == (0, 0, 0, 0, '')
    assert uwsgi.master_pid == -1
    assert uwsgi.config_variables == {}
    assert uwsgi.config == {}
    assert uwsgi.hostname == ''

    hooked = []

    @uwsgi.postfork_hooks.add()
    def hookit():
        hooked.append('yes')
        return True

    uwsgi.postfork_hooks.run()
    assert len(hooked) == 1
    assert hooked[0] == 'yes'



try:
    from django.urls import re_path

except ImportError:
    # fallback for Django < 4.0
    from django.conf.urls import re_path

from django.contrib import admin, messages
from django.template.response import TemplateResponse
from django.utils.translation import gettext_lazy as _


class OnePageAdmin(admin.ModelAdmin):

    def get_urls(self):
        meta = self.model._meta

        urlpatterns = [
            re_path(
                '^$',
                self.admin_site.admin_view(self.view_onepage),
                name=f'{meta.app_label}_{meta.model_name}_changelist'
            )
        ]

        return urlpatterns

    def view_onepage(self, request):

        # Not on the top to protect from module caching
        # caused by 1. import from manage command 2. from uWSGI
        from uwsgiconf import uwsgi

        is_stub = uwsgi.is_stub

        if is_stub:
            self.message_user(
                request,
                _("This site is not served by uWSGI. Certain functionality or data may be unavailable."),
                level=messages.WARNING,
            )

        context = dict(
            self.admin_site.each_context(request),

            stub=is_stub,
            title=self.opts.verbose_name,
            model_meta=self.model._meta,
        )
        self.contribute_onepage_context(request, context)
        return TemplateResponse(request, 'admin/uwsgify/onepage.html', context)

    def contribute_onepage_context(self, request, context):  # pragma: nocover
        raise NotImplementedError



import os

import pytest


@pytest.fixture
def patch_project_dir(monkeypatch):
    monkeypatch.setattr('uwsgiconf.contrib.django.uwsgify.toolbox.find_project_dir', lambda: 'dummy')


@pytest.fixture
def patch_base_command(monkeypatch, patch_project_dir, tmpdir):

    fifofile = tmpdir.join('some.fifo')
    fifofile.write('')

    monkeypatch.setattr(
        'uwsgiconf.contrib.django.uwsgify.toolbox.SectionMutator.get_fifo_filepath',
        lambda project_name: f'{fifofile}')


def test_mutate_existing_section(patch_base_command):
    from uwsgiconf.contrib.django.uwsgify.toolbox import SectionMutator

    mutator = SectionMutator.spawn(dir_base=os.path.dirname(__file__))
    assert mutator.section.name == 'testdummy'


def test_uwsgi_run(monkeypatch, patch_project_dir, command_run, settings, tmpdir, capsys):

    def runtime_dir(self):
        return f'{tmpdir}'

    monkeypatch.setattr('os.execvp', lambda *args, **kwargs: None)
    monkeypatch.setattr('uwsgiconf.contrib.django.uwsgify.toolbox.SectionMutator.runtime_dir', property(runtime_dir))

    with settings(STATIC_ROOT=f'{tmpdir}'):
        command_run('uwsgi_run')

    out, err = capsys.readouterr()
    assert tmpdir.join('admin').exists()
    assert tmpdir.join('uwsgify').exists()
    assert 'static files copied' in out

    command_run('uwsgi_run', options={'compile': True})
    out, err = capsys.readouterr()
    assert 'error-page-404 = replaceit/uwsgify/404.html' in out

    with pytest.raises(ImportError) as e:  # py3 - ModuleNotFoundError
        command_run('uwsgi_run', options={'embedded': True})
    out, err = capsys.readouterr()
    assert 'Deleting' in out
    assert f'{e.value}' == "No module named 'pyuwsgi'"


def test_uwsgi_log(patch_base_command, command_run):
    command_run('uwsgi_log', options={'reopen': True})


def test_uwsgi_reload(patch_base_command, command_run):
    command_run('uwsgi_reload')


def test_uwsgi_stats(patch_base_command, command_run):
    command_run('uwsgi_stats')


def test_uwsgi_stop(patch_base_command, command_run):
    command_run('uwsgi_stop')


def test_uwsgi_sysinit_systemd(patch_base_command, capsys, command_run):
    command_run('uwsgi_sysinit', options={'systype': 'systemd', 'nostatic': True, 'noruntimes': True})

    out, err = capsys.readouterr()

    uid = os.getuid()

    assert not err
    assert ('-o %s -g %s' % (uid, os.getgid())) in out
    assert ('/run/user/%s/dummy' % uid) in out
    assert 'Description=dummy uWSGI Service' in out
    assert 'bin/python' in out
    assert 'dummy/manage.py uwsgi_run --nostatic --noruntimes\n' in out


def test_uwsgi_sysinit_upstart(patch_base_command, capsys, command_run):
    command_run('uwsgi_sysinit', options={'systype': 'upstart', 'nostatic': True})

    out, err = capsys.readouterr()

    assert not err
    assert 'description "dummy uWSGI Service"' in out
    assert 'bin/python' in out
    assert 'dummy/manage.py uwsgi_run' in out


class TestAdmin:

    def test_configuration(self, request_client, user_create):

        client = request_client(user=user_create(superuser=True))
        data = client.get('/admin/uwsgify/configuration/').content.decode()
        assert 'This site is not served by uWSGI.' in data

    def test_summary(self, request_client, user_create):
        from uwsgiconf.runtime.signals import Signal

        signal = Signal()

        @signal.register_handler()
        def somefunc():
            pass

        client = request_client(user=user_create(superuser=True))
        data = client.get('/admin/uwsgify/summary/').content.decode()
        assert 'Requests total' in data
        assert '0 - worker: test_django.somefunc' in data

    def test_workers(self, request_client, user_create):

        client = request_client(user=user_create(superuser=True))
        data = client.get('/admin/uwsgify/workers/').content.decode()
        assert 'This site is not served by uWSGI.' in data

    def test_maintenance(self, request_client, user_create, monkeypatch, tmpdir):

        client = request_client(user=user_create(superuser=True))
        data = client.get('/admin/uwsgify/maintenance/').content.decode()
        assert 'This site is not served by uWSGI.' in data
        assert 'Maintenance mode is not supported' in data

        filepath = tmpdir.join('maintain')
        monkeypatch.setattr('uwsgiconf.settings.get_maintenance_path', lambda: f'{filepath}')

        data = client.get('/admin/uwsgify/maintenance/').content.decode()
        assert 'Remember: maintenance mode can only be' in data
        assert 'type="submit"' in data

        data = client.post('/admin/uwsgify/maintenance/', data={'confirm': 'bogus'}).content.decode()
        assert 'You need to enter' in data

        data = client.post('/admin/uwsgify/maintenance/', data={'confirm': 'maintenance'}).content.decode()
        assert 'Maintenance mode is scheduled' in data

        monkeypatch.setattr('uwsgiconf.settings.get_maintenance_path', lambda: f'/')
        data = client.post('/admin/uwsgify/maintenance/', data={'confirm': 'maintenance'}).content.decode()
        assert 'Unable to schedule' in data



from functools import wraps
from typing import Callable

from .. import uwsgi


class Lock:
    """Locks related stuff.

    Lock number 0 is always available. More locks need to be registered
    with ``.config.locking.set_basic_params(count=X)`` where ``X`` is the number of locks.

    .. note:: The same lock should be released before next acquiring.

    Can be used as context manager:

        .. code-block:: python

            with Lock():
                do()

    Can de used as a decorator:

        .. code-block:: python

            @Lock()
            def do():
                pass

    """
    __slots__ = ['num']

    def __init__(self, num: int = 0):
        """
        :param num: Lock number (0-64). 0 is always available and is used as default.

        """
        self.num = num

    def __int__(self):
        return self.num

    def __call__(self, func: Callable):

        @wraps(func)
        def wrapper(*args, **kwargs):

            with self:
                return func(*args, **kwargs)

        return wrapper

    @property
    def is_set(self) -> bool:
        """"Checks whether the lock is active.

        :raises ValueError: For Spooler or invalid lock number

        """
        return uwsgi.is_locked(self.num)

    def acquire(self):
        """Sets the lock.

        :raises ValueError: For Spooler or invalid lock number

        """
        uwsgi.lock(self.num)
        return True

    def release(self):
        """Unlocks the lock.

        :raises ValueError: For Spooler or invalid lock number

        """
        uwsgi.unlock(self.num)
        return True

    __enter__ = acquire

    def __exit__(self, exc_type, exc_value, traceback):
        self.release()


lock = Lock
"""Convenience alias for ``Lock``."""



import pytest

from uwsgiconf.config import Section
from uwsgiconf.exceptions import ConfigurationError


def test_networking_basics(assert_lines):

    assert_lines([
        'listen = 3',
    ], Section().networking.set_basic_params(queue_size=3))

    assert_lines([
        'so-keepalive = true',
    ], Section().networking.set_socket_params(keep_alive=True))

    assert_lines([
        'abstract-socket = true',
    ], Section().networking.set_unix_socket_params(abstract=True))

    assert_lines([
        'reuse-port = true',
    ], Section().networking.set_bsd_socket_params(port_reuse=True))

    sockets = Section.networking.sockets

    assert_lines([
        'http-socket = :8080',
        'map-socket = 0:1',
        'http-socket-modifier1 = 8',
        'http-socket-modifier2 = 1',

    ], Section().networking.register_socket(
        sockets.http(':8080', bound_workers=1, modifier=Section.routing.modifiers.jvm(submod=1))
    ))

    assert_lines([
        'socket = :8000',
        'map-socket = 0:2,3',
    ], Section().networking.register_socket(sockets.default(':8000', bound_workers=[2, 3])))

    assert_lines([
        'socket = /var/run/mine.sock',
    ], Section(
        runtime_dir='/var/run', project_name='mine'
    ).networking.register_socket(sockets.default('{project_runtime_dir}.sock')))

    assert_lines([
        'plugin = http',
        'https-export-cert = MYVAR',

    ], Section().networking.set_ssl_params(client_cert_var='MYVAR'))

    assert_lines([
        'http11-socket = :8001',
    ], Section().networking.register_socket(sockets.http(':8001', http11=True)))

    assert_lines([
        'https-socket = :8001,cert.crt,key.key,,!clientca',
    ], Section().networking.register_socket(
        sockets.https(
            ':8001', cert='cert.crt', key='key.key', client_ca='!clientca')))

    assert_lines([
        'puwsgi-socket = :8001',
    ], Section().networking.register_socket(sockets.uwsgi(':8001', persistent=True)))

    assert_lines([
        'fastcgi-nph-socket = :8001',
    ], Section().networking.register_socket(sockets.fastcgi(':8001', nph=True)))

    assert_lines([
        'scgi-nph-socket = :8001',
    ], Section().networking.register_socket(sockets.scgi(':8001', nph=True)))

    shared1 = sockets.shared(':80', undeferred=True)
    shared2 = sockets.shared(':81')

    assert_lines([
            'undeferred-shared-socket = :80\nsocket = =0\nuwsgi-socket = =0',
            'shared-socket = :81\nscgi-socket = =1',
        ],
        Section().networking.register_socket([
            sockets.default(shared1),
            sockets.uwsgi(shared1),
            sockets.scgi(shared2),
        ])
    )

    assert_lines([
            'shared-socket = :80\nshared-socket = :443',
            'http = =0\nhttps2 = cert=foobar.crt,key=foobar.key,addr==1',
        ],
        Section().routing.use_router(
            Section.routing.routers.http(sockets.shared(':80'))

        ).routing.use_router(
            Section.routing.routers.https(sockets.shared(':443'), cert='foobar.crt', key='foobar.key')
        )
    )


def test_networking_sockets(assert_lines):

    sockets = Section.networking.sockets

    socket = sockets.from_dsn('http://192.168.1.1:8000')
    assert isinstance(socket, sockets.http)
    assert str(socket) == '192.168.1.1:8000'

    socket = sockets.from_dsn('https://127.0.0.1:443?cert=/here/there.crt&key=/that/my.key')
    assert isinstance(socket, sockets.https)
    assert str(socket) == '127.0.0.1:443,/here/there.crt,/that/my.key'

    with pytest.raises(ConfigurationError):
        sockets.from_dsn('http://127.0.0.1:443?unsupported_arg=1020')


def test_networking_sni(assert_lines):

    assert_lines([
        'sni-regexp = *.pythonz.net /here/my.crt,/here/my.key,HIGH,/there/my.ca',

    ], Section().networking.set_sni_params(
        '*.pythonz.net', cert='/here/my.crt', key='/here/my.key',
        client_ca='/there/my.ca', ciphers='HIGH',
        wildcard=True)
    )

    assert_lines([
        'sni-dir = /certs/',
        'sni-dir-ciphers = MEDIUM',

    ], Section().networking.set_sni_dir_params('/certs/', ciphers='MEDIUM'))



from ._base import FifoCommand, Fifo


class Command(FifoCommand):

    help = 'Shutdown uWSGI instance'

    def add_arguments(self, parser):  # pragma: nocover

        super().add_arguments(parser)

        parser.add_argument(
            '--force', action='store_true', dest='force',
            help='Use forced (brutal) shutdown instead of a graceful one.',
        )

    def run_cmd(self, fifo: Fifo, options: dict):
        fifo.cmd_stop(force=options['force'])



import os

from .routing_actions import *
from .routing_modifiers import *
from .routing_routers import *
from .routing_subjects import *
from .routing_vars import *
from ..base import OptionsGroup
from ..exceptions import ConfigurationError
from ..utils import listify


class RouteRule:
    """Represents a routing rule."""

    class vars:
        """Routing variables."""

        cookie = VarCookie
        geoip = VarGeoip
        httptime = VarHttptime
        metric = VarMetric
        query = VarQuery
        request = VarRequest
        time = VarTime
        uwsgi = VarUwsgi

    class var_functions:
        """Functions that can be applied to variables."""

        base64 = FuncBase64
        hex = FuncHex
        lower = FuncLower
        math = FuncMath
        mime = FuncMime
        upper = FuncUpper

    class stages:
        """During the request cycle, various stages (aka chains) are processed.

        Chains can be "recursive". A recursive chain can be called multiple times
        in a request cycle.

        """

        REQUEST = ''
        """Applied before the request is passed to the plugin."""

        ERROR = 'error'
        """Applied as soon as an HTTP status code is generate. **Recursive chain**."""

        RESPONSE = 'response'
        """Applied after the last response header has been generated (just before sending the body)."""

        FINAL = 'final'
        """Applied after the response has been sent to the client."""

    class subjects:
        """Routing subjects. These can be request's variables or other entities.

        .. note:: Non-custom subjects can be pre-optimized (during startup)
            and should be used for performance reasons.

        """
        custom = SubjectCustom

        http_host = SubjectHttpHost
        http_referer = SubjectHttpReferer
        http_user_agent = SubjectHttpUserAgent
        path_info = SubjectPathInfo
        query_string = SubjectQueryString
        remote_addr = SubjectRemoteAddr
        remote_user = SubjectRemoteUser
        request_uri = SubjectRequestUri
        status = SubjectStatus

    class transforms:
        """A transformation is like a filter applied to the response
        generated by your application.

        Transformations can be chained (the output of a transformation will be the input of the following one)
        and can completely overwrite response headers.

        * http://uwsgi.readthedocs.io/en/latest/Transformations.html

        """
        chunked = ActionChunked
        fix_content_len = ActionFixContentLen
        flush = ActionFlush
        gzip = ActionGzip
        template = ActionTemplate
        to_file = ActionToFile
        upper = ActionUpper

        # todo Consider adding the following and some others from sources (incl. plugins):
        # xslt, cachestore, memcachedstore, redisstore, rpc, lua

    class actions:
        """Actions available for routing rules.

        Values returned by actions:

            * ``NEXT`` - continue to the next rule
            * ``CONTINUE`` - stop scanning the internal routing table and run the request
            * ``BREAK`` - stop scanning the internal routing table and close the request
            * ``GOTO x`` - go to rule ``x``

        """
        add_var_cgi = ActionAddVarCgi
        add_var_log = ActionAddVarLog
        alarm = ActionAlarm
        auth_basic = ActionAuthBasic
        auth_ldap = AuthLdap
        dir_change = ActionDirChange
        do_break = ActionDoBreak
        do_continue = ActionDoContinue
        do_goto = ActionDoGoto
        fix_var_path_info = ActionFixVarPathInfo
        header_add = ActionHeaderAdd
        header_remove = ActionHeaderRemove
        headers_off = ActionHeadersOff
        headers_reset = ActionHeadersReset
        log = ActionLog
        offload_off = ActionOffloadOff
        redirect = ActionRedirect
        rewrite = ActionRewrite
        route_external = ActionRouteExternal
        route_uwsgi = ActionRouteUwsgi
        send = ActionSend
        serve_static = ActionServeStatic
        set_harakiri = ActionSetHarakiri
        set_script_file = ActionSetScriptFile
        set_uwsgi_process_name = ActionSetUwsgiProcessName
        set_var_document_root = ActionSetVarDocumentRoot
        set_var_path_info = ActionSetVarPathInfo
        set_var_remote_addr = ActionSetVarRemoteAddr
        set_var_remote_user = ActionSetVarRemoteUser
        set_var_request_method = ActionSetVarRequestMethod
        set_var_request_uri = ActionSetVarRequestUri
        set_var_script_name = ActionSetVarScriptName
        set_var_uwsgi_appid = ActionSetVarUwsgiAppid
        set_var_uwsgi_home = ActionSetVarUwsgiHome
        set_var_uwsgi_scheme = ActionSetVarUwsgiScheme
        signal = ActionSignal

        # todo Consider adding the following and some others from sources (incl. plugins):
        # cachestore, cacheset, memcached,
        # router_cache: cache, cache-continue, cachevar, cacheinc, cachedec, cachemul, cachediv
        # rpc,
        # rpc: call, rpcret, rpcnext, rpcraw, rpcvar,
        # access, spnego, radius
        # xslt, ssi, gridfs
        # cgi: cgi, cgihelper
        # router_access: access,
        # proxyhttp -router_http, proxyuwsgi -router_uwsgi, xattr -xattr
        # router_memcached: memcached, memcached-continue, memcachedstore
        # router_redis: redis, redis-continue, redisstore

    def __init__(self, action, subject=None, stage=stages.REQUEST):
        """
        :param RouteAction action: Action (or transformation) to perfrom.
            See ``.actions`` and ``.transforms``.

        :param SubjectCustom|SubjectBuiltin|str subject: Subject to verify before action is performed.
            See ``.subjects``.

            * String values are automatically transformed into ``subjects.path_info``.
            * If ``None`` action is performed always w/o subject check.

        :param str stage: Stage on which the action needs to be performed.
            See ``.stages``.

        """
        if subject is None:
            subject = 'run'  # always run the specified route action

        elif isinstance(subject, str):
            subject = self.subjects.path_info(subject)

        subject_rule = ''

        self._custom_subject = isinstance(subject, SubjectCustom)

        if self._custom_subject:
            subject_rule = subject
            subject = 'if-not' if subject.negate else 'if'

        elif isinstance(subject, SubjectBuiltin):
            subject_rule = subject.regexp
            subject = subject.name

        self.command_label = f'{stage}-route-label'.strip('-')
        self.command = f'{stage}-route-{subject}'.strip('-')
        self.value = subject_rule, action


class Routing(OptionsGroup):
    """Routing subsystem.

    You can use the internal routing subsystem to dynamically alter the way requests are handled.

    .. note:: Since 1.9

    * http://uwsgi.readthedocs.io/en/latest/InternalRouting.html
    * http://uwsgi.readthedocs.io/en/latest/Transformations.html

    """

    route_rule = RouteRule

    class routers:
        """Dedicated routers, which can be used with `register_router()`."""

        http = RouterHttp
        https = RouterHttps
        ssl = RouterSsl
        fast = RouterFast
        raw = RouterRaw
        forkpty = RouterForkPty
        tuntap = RouterTunTap

    class modifiers:
        """Routing modifiers.

        * http://uwsgi.readthedocs.io/en/latest/Protocol.html

        """
        cache = ModifierCache
        cgi = ModifierCgi
        cluster_node = ModifierClusterNode
        config_from_node = ModifierConfigFromNode
        corerouter_signal = ModifierCorerouterSignal
        echo = ModifierEcho
        eval = ModifierEval
        example = ModifierExample
        fastfunc = ModifierFastfunc
        gccgo = ModifierGccgo
        glusterfs = ModifierGlusterfs
        gridfs = ModifierGridfs
        jvm = ModifierJvm
        legion_msg = ModifierLegionMsg
        lua = ModifierLua
        manage = ModifierManage
        manage_path_info = ModifierManagePathInfo
        message = ModifierMessage
        message_array = ModifierMessageArray
        message_marshal = ModifierMessageMarshal
        mono = ModifierMono
        multicast = ModifierMulticast
        multicast_announce = ModifierMulticastAnnounce
        persistent_close = ModifierPersistentClose
        php = ModifierPhp
        ping = ModifierPing
        psgi = ModifierPsgi
        rack = ModifierRack
        rados = ModifierRados
        raw = ModifierRaw
        reload = ModifierReload
        reload_brutal = ModifierReloadBrutal
        remote_logging = ModifierRemoteLogging
        response = ModifierResponse
        rpc = ModifierRpc
        signal = ModifierSignal
        snmp = ModifierSnmp
        spooler = ModifierSpooler
        ssi = ModifierSsi
        subscription = ModifierSubscription
        symcall = ModifierSymcall
        v8 = ModifierV8
        webdav = ModifierWebdav
        wsgi = ModifierWsgi
        xslt = ModifierXslt

    def use_router(self, router, *, force=None):
        """

        :param RouterBase router: Dedicated router object. See `.routers`.

        :param bool force: All of the gateways (routers) has to be run under the master process,
            supplying this you can try to bypass this limit.

        """
        self._set('force-gateway', force, cast=bool)

        router._contribute_to_opts(self)

        return self._section

    def register_route(self, route_rules, *, label=None):
        """Registers a routing rule.

        :param RouteRule|list[RouteRule] route_rules:

        :param str label: Label to mark the given set of rules.
            This can be used in conjunction with ``do_goto`` rule action.

            * http://uwsgi.readthedocs.io/en/latest/InternalRouting.html#goto

        """
        route_rules = listify(route_rules)

        if route_rules and label:
            self._set(route_rules[0].command_label, label, multi=True)

        for route_rules in route_rules:
            self._set(route_rules.command, route_rules.value, multi=True)

        return self._section

    def print_routing_rules(self):
        """Print out supported routing rules (actions, transforms, etc.)."""

        self._set('routers-list', True, cast=bool)

        return self._section

    def set_error_page(self, status: int, html_fpath: str):
        """Add an error page (html) for managed 403, 404, 500 response.

        :param status: HTTP status code.

        :param html_fpath: HTML page file path.

        """
        statuses = [403, 404, 500]

        status = int(status)

        if status not in statuses:
            raise ConfigurationError(
                f"Code `{status}` for `routing.set_error_page()` is unsupported. "
                f"Supported: {', '.join(map(str, statuses))}")

        self._set(f'error-page-{status}', html_fpath, multi=True)

        return self._section

    def set_error_pages(self, codes_map: dict = None, *, common_prefix: str = None):
        """Add an error pages for managed 403, 404, 500 responses.

        Shortcut for ``.set_error_page()``.

        :param codes_map: Status code mapped into an html filepath or
            just a filename if common_prefix is used.

            If not set, filename containing status code is presumed: 400.html, 500.html, etc.

        :param common_prefix: Common path (prefix) for all files.

        """
        statuses = [403, 404, 500]

        if common_prefix:
            if not codes_map:
                codes_map = {code: f'{code}.html' for code in statuses}

            for code, filename in codes_map.items():
                codes_map[code] = os.path.join(common_prefix, filename)

        for code, filepath in codes_map.items():
            self.set_error_page(code, filepath)

        return self._section

    def set_geoip_params(self, *, db_country=None, db_city=None):
        """Sets GeoIP parameters.

        * http://uwsgi.readthedocs.io/en/latest/GeoIP.html

        :param str db_country: Country database file path.

        :param str db_city: City database file path. Example: ``GeoLiteCity.dat``.

        """
        self._set('geoip-country', db_country, plugin='geoip')
        self._set('geoip-city', db_city, plugin='geoip')

        return self._section

    def header_add(self, name, value):
        """Automatically add HTTP headers to response.

        :param str name:

        :param str value:

        """
        self._set('add-header', f'{name}: {value}', multi=True)

        return self._section

    def header_remove(self, value):
        """Automatically remove specified HTTP header from the response.

        :param str value:

        """
        self._set('del-header', value, multi=True)

        return self._section

    def header_collect(self, name, target_var, *, pull=False):
        """Store the specified response header in a request var
        (optionally removing it from the response).

        :param str name:

        :param str target_var:

        :param bool pull: Whether to remove header from response.

        """
        self._set(
            'pull-header' if pull else 'collect-header',
            f'{name} {target_var}',
            multi=True
        )

        return self._section



import sys
from contextlib import contextmanager

import click

from uwsgiconf import VERSION
from uwsgiconf.exceptions import ConfigurationError
from uwsgiconf.sysinit import TYPES, TYPE_SYSTEMD, get_config
from uwsgiconf.utils import ConfModule, UwsgiRunner


@contextmanager
def errorprint():
    """Print out descriptions from ConfigurationError."""
    try:
        yield

    except ConfigurationError as e:
        click.secho(f'{e}', err=True, fg='red')
        sys.exit(1)


@click.group()
@click.version_option(version='.'.join(map(str, VERSION)))
def base():
    """uwsgiconf command line utility.

    Tools to facilitate uWSGI configuration.

    """

arg_conf = click.argument(
    'conf', type=click.Path(exists=True, dir_okay=False), default=ConfModule.default_name)


@base.command()
@arg_conf
@click.option('--only', help='Configuration alias from module to run uWSGI with.')
def run(conf, only):
    """Runs uWSGI passing to it using the default or another `uwsgiconf` configuration module.

    """
    with errorprint():
        config = ConfModule(conf)
        spawned = config.spawn_uwsgi(only=only)

        for alias, pid in spawned:
            click.secho(f"Spawned uWSGI for configuration aliased '{alias}'. PID {pid}", fg='green')


@base.command()
@arg_conf
def compile(conf):
    """Compiles classic uWSGI configuration file using the default
    or given `uwsgiconf` configuration module.

    """
    with errorprint():
        config = ConfModule(conf)
        for conf in config.configurations:
            conf.format(do_print=True)


@base.command()
@click.argument('systype', type=click.Choice(TYPES), default=TYPE_SYSTEMD)
@arg_conf
@click.option('--project', help='Project name to use as service name.')
def sysinit(systype, conf, project):
    """Outputs configuration for system initialization subsystem."""

    click.secho(get_config(
        systype,
        conf=ConfModule(conf).configurations[0],
        conf_path=conf,
        project_name=project,
    ))


@base.command()
def probe_plugins():
    """Runs uWSGI to determine what plugins are available and prints them out.

    Generic plugins come first then after blank line follow request plugins.

    """
    plugins = UwsgiRunner().get_plugins()

    for plugin in sorted(plugins.generic):
        click.secho(plugin)

    click.secho('')

    for plugin in sorted(plugins.request):
        click.secho(plugin)


def main():
    """
    CLI entry point
    """
    base(obj={})


if __name__ == '__main__':
    main()



from ._base import FifoCommand, Fifo


class Command(FifoCommand):

    help = 'Reloads uWSGI master process, workers'

    def add_arguments(self, parser):  # pragma: nocover

        super().add_arguments(parser)

        parser.add_argument(
            '--force', action='store_true', dest='force',
            help='Use forced (brutal) reload instead of a graceful one.',
        )
        parser.add_argument(
            '--workers-only', action='store_true', dest='workers',
            help='Reload only workers.',
        )
        parser.add_argument(
            '--workers-chain', action='store_true', dest='chain',
            help='Run chained workers reload (one after another, instead of destroying all of them in bulk).',
        )

    def run_cmd(self, fifo: Fifo, options: dict):
        fifo.cmd_reload(
            force=options['force'],
            workers_only=options['workers'],
            workers_chain=options['chain'],
        )



from os import environ

CONFIGS_MODULE_ATTR = 'uwsgi_configuration'

ENV_CONF_ALIAS = 'UWSGICONF_CONF_ALIAS'
ENV_CONF_READY = 'UWSGICONF_READY'
ENV_FORCE_STUB = 'UWSGICONF_FORCE_STUB'
ENV_MAINTENANCE = 'UWSGICONF_MAINTENANCE'
ENV_MAINTENANCE_INPLACE = 'UWSGICONF_MAINTENANCE_INPLACE'


FORCE_STUB = int(environ.get(ENV_FORCE_STUB, 0))
"""Forces using stub instead of uwsgi real module."""


def get_maintenance_path() -> str:
    """Return the maintenance trigger filepath.
    Introduced as a function to support embedded mode.

    """
    return environ.get(ENV_MAINTENANCE) or ''


def get_maintenance_inplace() -> bool:
    """Return the maintenance flag if it is set.
    Introduced as a function to support embedded mode.

    """
    return int(environ.get(ENV_MAINTENANCE_INPLACE, 0)) != 0



from pathlib import Path
from typing import Union, Callable, Optional

from ..config import Section as _Section
from ..settings import ENV_MAINTENANCE, ENV_MAINTENANCE_INPLACE
from ..typehints import Strlist


class Section(_Section):
    """Basic nice configuration."""

    def __init__(
            self,
            name: str = None,
            *,
            touch_reload: Strlist =None,
            workers: int = None,
            threads: Union[int, bool] = None,
            mules: int = None,
            owner: str = None,
            log_into: str = None,
            log_dedicated: bool = None,
            process_prefix: str = None,
            ignore_write_errors: bool = None,
            **kwargs):
        """

        :param name: Section name.

        :param touch_reload: Reload uWSGI if the specified file or directory is modified/touched.

        :param workers: Spawn the specified number of workers (processes).
            Default: workers number equals to CPU count.

        :param threads: Number of threads per worker or ``True`` to enable user-made threads support.

        :param mules: Number of mules to spawn.

        :param owner: Set process owner user and group.

        :param log_into: Filepath or UDP address to send logs into.

        :param log_dedicated: If ``True`` all logging will be handled with a separate
            thread in master process.

        :param process_prefix: Add prefix to process names.

        :param ignore_write_errors: If ``True`` no annoying SIGPIPE/write/writev errors
            will be logged, and no related exceptions will be raised.

            .. note:: Usually such errors could be seen on client connection cancelling
               and are safe to ignore.

        :param kwargs:

        """
        super().__init__(strict_config=True, name=name, **kwargs)

        # Fix possible problems with non-ASCII.
        self.env('LANG', 'en_US.UTF-8')

        if touch_reload:
            self.main_process.set_basic_params(touch_reload=touch_reload)

        if workers:
            self.workers.set_basic_params(count=workers)
        else:
            self.workers.set_count_auto()

        set_threads = self.workers.set_thread_params

        if isinstance(threads, bool):
            set_threads(enable=threads)

        else:
            set_threads(count=threads)

        if log_dedicated:
            self.logging.set_master_logging_params(enable=True, dedicate_thread=True)

        self.workers.set_mules_params(mules=mules)
        self.workers.set_harakiri_params(verbose=True)
        self.main_process.set_basic_params(vacuum=True)
        self.main_process.set_naming_params(
            autonaming=True,
            prefix=f'{process_prefix} ' if process_prefix else None,
        )
        self.master_process.set_basic_params(enable=True)
        self.master_process.set_exit_events(sig_term=True)  # Respect the convention. Make Upstart and Co happy.
        self.locks.set_basic_params(thunder_lock=True)
        self.configure_owner(owner=owner)
        self.logging.log_into(target=log_into)

        if ignore_write_errors:
            self.master_process.set_exception_handling_params(no_write_exception=True)
            self.logging.set_filters(write_errors=False, sigpipe=False)

    def get_log_format_default(self) -> str:
        """Returns default log message format.

        .. note:: Some params may be missing.

        """
        vars = self.logging.vars

        app_id = app_req_count = worker_req_count = tsize = sendfile_route_offload = '-'

        format_default = (
            f"[pid: {vars.WORKER_PID}|app: {app_id}|req: {app_req_count}/{worker_req_count}] "
            f"{vars.REQ_REMOTE_ADDR} ({vars.REQ_REMOTE_USER}) "
            f"{{{vars.REQ_COUNT_VARS_CGI} vars in {vars.SIZE_PACKET_UWSGI} bytes}} "
            f"[{vars.REQ_START_CTIME}] {vars.REQ_METHOD} {vars.REQ_URI} => "
            f"generated {vars.RESP_SIZE_BODY} bytes in {vars.RESP_TIME_MS} "
            f"{tsize}{sendfile_route_offload}({vars.REQ_SERVER_PROTOCOL} {vars.RESP_STATUS}) "
            f"{vars.RESP_COUNT_HEADERS} headers in {vars.RESP_SIZE_HEADERS} bytes "
            f"({vars.ASYNC_SWITCHES} switches on core {vars.CORE})"
        )
        return format_default

    @classmethod
    def get_bundled_static_path(cls, filename: str) -> str:
        """Returns a full path to a static HTML page file bundled with uwsgiconf.

        :param filename: File name to return path to.

            Examples:
                * 403.html
                * 404.html
                * 500.html
                * 503.html

        """
        return str(Path(__file__).parent.parent / 'contrib/django/uwsgify/static/uwsgify' / filename)

    def configure_maintenance_mode(self, trigger: Union[str, Path], response: str):
        """Allows maintenance mode when a certain response
        is given for every request if a trigger is set.

        :param trigger: This triggers maintenance mode responses.
            Should be a path to a file: if file exists, maintenance mode is on.

        :param response: Response to to give in maintenance mode.

            Supported:
                1. File path - this file will be served in response.

                2. URLs starting with ``http`` - requests will be redirected there using 302.
                  This is often discouraged, because it may have search ranking implications.

                3. Prefix ``app`` will replace your entire app with a maintenance one.
                  Using this also prevents background tasks registration and execution
                  (including scheduler, timers, signals).

                  * If the value is `app` - the default maintenance application bundled with
                    uwsgiconf would be used.

                  * Format ``app::<your-module>:<your-app-function>`` instructs uwsgiconf to
                    load your function as a maintenance app. E.g.: app::my_pack.my_module:my_func

        """
        if response.startswith('app'):

            if response == 'app':
                response = 'uwsgiconf.maintenance:app_maintenance'

            else:
                _, _, response = response.partition('::')

            filepath = Path(trigger)

            # We reload uwsgi to swap our app for maintenance app.
            self.main_process.set_basic_params(touch_reload=f'{filepath}')

            # This will also make available Maintenance page in Django admin.
            self.env(ENV_MAINTENANCE, f'{filepath}', update_local=True)

            if filepath.exists():
                self.env(ENV_MAINTENANCE_INPLACE, 1, update_local=True)
                self.python.set_wsgi_params(module=response)

            return self

        routing = self.routing

        rule = routing.route_rule

        if response.startswith('http'):
            action = rule.actions.redirect(response)

        else:
            action = rule.actions.serve_static(Path(response).absolute())

        routing.register_route(
            routing.route_rule(
                subject=rule.subjects.custom(trigger).exists(),
                action=action,
            ))

        return self

    def configure_owner(self, owner: str = 'www-data'):
        """Shortcut to set process owner data.

        :param owner: Sets user and group. Default: ``www-data``.

        """
        if owner is not None:
            self.main_process.set_owner_params(uid=owner, gid=owner)

        return self

    def configure_https_redirect(self):
        """Enables HTTP to HTTPS redirect."""
        rule = self.routing.route_rule

        self.routing.register_route([
            rule(
                rule.actions.redirect('https://${HTTP_HOST}${REQUEST_URI}', permanent=True),
                rule.subjects.custom('${HTTPS}', negate=True).eq('on')
            ),
        ])
        return self

    def configure_certbot_https(
            self,
            domain: str,
            webroot: str,
            *,
            address: str = None,
            allow_shared_sockets: bool = None,
            http_redirect: bool = False,
    ):
        """Enables HTTPS using certificates from Certbot https://certbot.eff.org.

        .. note:: This relies on ``webroot`` mechanism of Certbot - https://certbot.eff.org/docs/using.html#webroot

            Sample command to get a certificate: ``certbot certonly --webroot -w /webroot/path/ -d mydomain.org``

        :param domain: Domain name certificates issued for
            (the same as in ``-d`` option in the above command).

        :param webroot: Directory to store challenge files to get and renew the certificate
            (the same as in ``-w`` option in the above command).

        :param address: Address to bind socket to.

        :param allow_shared_sockets: Allows using shared sockets to bind
            to privileged ports. If not provided automatic mode is enabled:
            shared are allowed if current user is not root.

        :param http_redirect: Redirect HTTP requests to HTTPS
            if certificates exist.

        """
        address = address or ':443'

        networking = self.networking

        path_cert_chain, path_cert_private = networking.sockets.https.get_certbot_paths(domain)

        if path_cert_chain:

            networking.register_socket(
                networking.sockets.from_dsn(
                    f'https://{address}?cert={path_cert_chain}&key={path_cert_private}',
                    allow_shared_sockets=allow_shared_sockets))

            if http_redirect:
                self.configure_https_redirect()

        self.statics.register_static_map(
            '/.well-known/', webroot, retain_resource_path=True)

        return self

    def configure_logging_json(self):
        """Configures uWSGI output to be json-formatted."""

        logging = self.logging

        vars_enc = logging.encoders.format.vars
        vars_req = logging.vars

        # Allow request encoding.
        logging.add_logger(logging.loggers.stdio(), requests_only=True)

        # Log essential request data to place into "msg".
        logging.set_basic_params(template=(
            f'{vars_req.REQ_METHOD} {vars_req.REQ_URI} -> {vars_req.RESP_STATUS}'
        ))

        new_line = logging.encoders.newline()

        log_template = (
            '{'
            '"dt": "' + str(vars_enc.TIME_FORMAT('iso')) + '", '
            '"src": "__src__", '
            f'"msg": "{vars_enc.MESSAGE}", '
            '"ctx": {}, '
            f'"ms": {vars_enc.TIME_MS}'
            '}'
        )

        logging.add_logger_encoder(
            [logging.encoders.json(log_template.replace('__src__', 'uwsgi.req')), new_line],
            requests_only=True)

        logging.add_logger_encoder(
            [logging.encoders.json(log_template.replace('__src__', 'uwsgi.out')), new_line])

        return self


class PythonSection(Section):
    """Basic nice configuration using Python plugin."""

    def __init__(
            self,
            name: str = None,
            *,
            params_python: dict = None,
            wsgi_module: str = None,
            wsgi_callable: Union[str, Callable] = None,
            embedded_plugins: Optional[bool] = True,
            require_app: bool = True,
            threads: Union[bool, int] = True,
            **kwargs
    ):
        """

        :param name: Section name.

        :param params_python: See Python plugin basic params.

        :param wsgi_module: WSGI application module path or filepath.

            Example:
                mypackage.my_wsgi_module -- read from `application` attr of mypackage/my_wsgi_module.py
                mypackage.my_wsgi_module:my_app -- read from `my_app` attr of mypackage/my_wsgi_module.py

        :param wsgi_callable: WSGI application callable name. Default: application.

        :param embedded_plugins: This indicates whether plugins were embedded into uWSGI,
            which is the case if you have uWSGI from PyPI.

        :param require_app: Exit if no app can be loaded.

        :param threads: Number of threads per worker or ``True`` to enable user-made threads support.

        :param kwargs:

        """
        if embedded_plugins is True:
            embedded_plugins = self.embedded_plugins_presets.BASIC + ['python', 'python2', 'python3']

        super().__init__(
            name=name, embedded_plugins=embedded_plugins, threads=threads,
            **kwargs)

        self.python.set_basic_params(**(params_python or {}))

        if callable(wsgi_callable):
            wsgi_callable = wsgi_callable.__name__

        self.python.set_wsgi_params(module=wsgi_module, callable_name=wsgi_callable)

        self.applications.set_basic_params(exit_if_none=require_app)



import subprocess
from os import environ

import pytest

from uwsgiconf.settings import ENV_FORCE_STUB

# Force stub to allow shallow testing.
environ[ENV_FORCE_STUB] = '1'

from pytest_djangoapp import configure_djangoapp_plugin


pytest_plugins = configure_djangoapp_plugin(
    {
        'ROOT_URLCONF': 'tests.contrib.urls',
        'WSGI_APPLICATION': 'tests.contrib.app',
        'STATIC_ROOT': 'replaceit/',
    },
    app_name='uwsgiconf.contrib.django.uwsgify',
    extend_MIDDLEWARE=[
        # messages in admin support
        'django.contrib.sessions.middleware.SessionMiddleware',
        'django.contrib.messages.middleware.MessageMiddleware',
    ],
    extend_INSTALLED_APPS=[
        'django.contrib.staticfiles',
    ],
    admin_contrib=True,
)


@pytest.fixture(scope='session')
def ini():

    def wrapped(section, stamp):
        return section.as_configuration().format(stamp=stamp)

    return wrapped


@pytest.fixture(scope='session')
def assert_lines(ini):

    def wrapped(lines, source, assert_in=True, stamp=False):
        source = ini(source, stamp)

        if not isinstance(lines, list):
            lines = [lines]

        for line in lines:
            if assert_in:
                assert line in source, source
            else:
                assert line not in source, source

    return wrapped


@pytest.fixture
def mock_popen(monkeypatch):

    def mock(func_communicate):

        class MockPopen(object):

            def __init__(self, *args, **kwargs):
                pass

            def communicate(self, *args, **kwargs):
                return func_communicate(*args, **kwargs)

        monkeypatch.setattr(subprocess, 'Popen', MockPopen)

    return mock



from os import geteuid
from urllib.parse import urlsplit, parse_qs

from .networking_sockets import *
from ..base import OptionsGroup
from ..exceptions import ConfigurationError


class Networking(OptionsGroup):
    """Networking related stuff. Socket definition, binding and tuning."""
    
    class sockets:
        """Available socket types to use with ``.register_socket()``."""

        default = SocketDefault
        fastcgi = SocketFastcgi
        http = SocketHttp
        https = SocketHttps
        raw = SocketRaw
        scgi = SocketScgi
        shared = SocketShared
        udp = SocketUdp
        uwsgi = SocketUwsgi
        uwsgis = SocketUwsgis
        zeromq = SocketZeromq

        @classmethod
        def from_dsn(cls, dsn, *, allow_shared_sockets=None) -> 'Socket':
            """Constructs socket configuration object from DSN.

            .. note:: This will also automatically use shared sockets
                to bind to priviledged ports when non root.

            :param str dsn: Data source name, e.g:
                * http://127.0.0.1:8000
                * https://127.0.0.1:443?cert=/here/there.crt&key=/that/my.key

                .. note:: Some schemas:
                    fastcgi, http, https, raw, scgi, shared, udp, uwsgi, suwsgi, zeromq

            :param bool allow_shared_sockets: Allows using shared sockets to bind
                to priviledged ports. If not provided automatic mode is enabled:
                shared are allowed if current user is not root.

            :rtype Socket

            """
            split = urlsplit(dsn)

            sockets = {
                socket.name.replace('socket', '').rstrip('-'): socket
                for socket in cls.__dict__.values() if isinstance(socket, type) and issubclass(socket, Socket)}

            socket_kwargs = {
                'address': split.netloc,
            }
            socket_kwargs.update({key: val[0] for key, val in parse_qs(split.query).items()})
            socket = sockets[split.scheme]

            if split.port and split.port < 1024:

                if allow_shared_sockets is None:
                    allow_shared_sockets = (geteuid() != 0)

                if allow_shared_sockets:
                    new_shared = cls.shared(socket_kwargs['address'])
                    socket_kwargs['address'] = new_shared

            try:
                socket = socket(**socket_kwargs)

            except TypeError as e:
                raise ConfigurationError(
                    f'Unable to configure {socket.__name__} using `{dsn}` DSN: {e}')

            return socket

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._sockets = []  # Registered sockets list.

    def set_basic_params(self, *, queue_size=None, freebind=None, default_socket_type=None):
        """

        :param int queue_size: Also known as a backlog. Every socket has an associated queue 
            where request will be put waiting for a process to became ready to accept them. 
            When this queue is full, requests will be rejected.
            
            Default: 100 (an average value chosen by the maximum value allowed by default 
            by your kernel).

            .. note:: The maximum value is system/kernel dependent. Before increasing it you may 
                need to increase your kernel limit too.

        :param bool freebind: Put socket in freebind mode.
            Allows binding to non-existent network addresses.

            .. note:: Linux only.

        :param str default_socket_type: Force the socket type as default.
            See ``.socket_types``.

        """
        self._set('listen', queue_size)
        self._set('freebind', freebind, cast=bool)
        self._set('socket-protocol', default_socket_type)

        return self._section

    def set_socket_params(
            self, *, send_timeout=None, keep_alive=None, no_defer_accept=None,
            buffer_send=None, buffer_receive=None):
        """Sets common socket params.

        :param int send_timeout: Send (write) timeout in seconds.

        :param bool keep_alive: Enable TCP KEEPALIVEs.

        :param bool no_defer_accept: Disable deferred ``accept()`` on sockets
            by default (where available) uWSGI will defer the accept() of requests until some data
            is sent by the client (this is a security/performance measure).
            If you want to disable this feature for some reason, specify this option.

        :param int buffer_send: Set SO_SNDBUF (bytes).

        :param int buffer_receive: Set SO_RCVBUF (bytes).

        """
        self._set('so-send-timeout', send_timeout)
        self._set('so-keepalive', keep_alive, cast=bool)
        self._set('no-defer-accept', no_defer_accept, cast=bool)
        self._set('socket-sndbuf', buffer_send)
        self._set('socket-rcvbuf', buffer_receive)

        return self._section

    def set_unix_socket_params(self, *, abstract=None, permissions=None, owner=None, umask=None):
        """Sets Unix-socket related params.

        :param bool abstract: Force UNIX socket into abstract mode (Linux only).

        :param str permissions: UNIX sockets are filesystem objects that obey
            UNIX permissions like any other filesystem object.

            You can set the UNIX sockets' permissions with this option if your webserver
            would otherwise have no access to the uWSGI socket. When used without a parameter,
            the permissions will be set to 666. Otherwise the specified chmod value will be used.

        :param str owner: Chown UNIX sockets.

        :param str umask: Set UNIX socket umask.

        """
        self._set('abstract-socket', abstract, cast=bool)
        self._set('chmod-socket', permissions)
        self._set('chown-socket', owner)
        self._set('umask', umask)

        return self._section

    def set_bsd_socket_params(self, *, port_reuse=None):
        """Sets BSD-sockets related params.

        :param bool port_reuse: Enable REUSE_PORT flag on socket to allow multiple
            instances binding on the same address (BSD only).

        """
        self._set('reuse-port', port_reuse, cast=bool)

        return self._section

    def _get_shared_socket_idx(self, shared):
        return f'={self._sockets.index(shared)}'

    def register_socket(self, socket):
        """Registers the given socket(s) for further use.

        :param Socket|list[Socket] socket: Socket type object. See ``.sockets``.

        """
        sockets = self._sockets

        for socket in listify(socket):

            uses_shared = isinstance(socket.address, SocketShared)

            if uses_shared:
                # Handling shared sockets involves socket index resolution.

                shared_socket: SocketShared = socket.address

                if shared_socket not in sockets:
                    self.register_socket(shared_socket)

                socket.address = self._get_shared_socket_idx(shared_socket)

            socket.address = self._section.replace_placeholders(socket.address)
            self._set(socket.name, socket, multi=True)

            socket._contribute_to_opts(self)

            bound_workers = socket.bound_workers

            if bound_workers:
                self._set(
                    'map-socket', f"{len(sockets)}:{','.join(map(str, bound_workers))}",
                    multi=True)

            if not uses_shared:
                sockets.append(socket)

        return self._section

    def set_ssl_params(
            self, *, verbose_errors=None,
            sessions_cache=None, sessions_timeout=None, session_context=None,
            raw_options=None, dir_tmp=None, client_cert_var=None):
        """

        :param bool verbose_errors: Be verbose about SSL errors.

        :param str|bool sessions_cache: Use uWSGI cache for ssl sessions storage.

            Accepts either bool or cache name string.

            * http://uwsgi.readthedocs.io/en/latest/SSLScaling.html

            .. warning:: Please be sure to configure cache before setting this.

        :param int sessions_timeout: Set SSL sessions timeout in seconds. Default: 300.

        :param str session_context: Session context identifying string. Can be set to static shared value
            to avoid session rejection.

            Default: a value built from the HTTP server address.

            * http://uwsgi.readthedocs.io/en/latest/SSLScaling.html#setup-2-synchronize-caches-of-different-https-routers

        :param int|list[int] raw_options: Set a raw ssl option by its numeric value.

        :param str dir_tmp: Store ssl-related temp files (e.g. pem data) in the specified directory.

        :param str client_cert_var: Export uWSGI variable ``HTTPS_CC`` containing the raw client certificate.

        """
        self._set('ssl-verbose', verbose_errors, cast=bool)
        self._set('ssl-sessions-use-cache', sessions_cache, cast=bool if isinstance(sessions_cache, bool) else None)
        self._set('ssl-sessions-timeout', sessions_timeout)

        for option in listify(raw_options):
            self._set('ssl-option', option, multi=True)

        self._set('ssl-tmp-dir', dir_tmp)

        self._set('https-session-context', session_context, plugin='http')
        self._set('https-export-cert', client_cert_var, plugin='http')

        return self._section

    def set_sni_params(
            self,
            name: str,
            *,
            cert: str,
            key: str,
            ciphers: str = None,
            client_ca: str = None,
            wildcard: bool = False
    ):
        """Allows setting Server Name Identification (virtual hosting for SSL nodes) params.

        * http://uwsgi.readthedocs.io/en/latest/SNI.html

        :param name: Node/server/host name.

        :param cert: Certificate file.

        :param key: Private key file.

        :param ciphers: Ciphers [alias] string.

            Example:
                * DEFAULT
                * HIGH
                * DHE, EDH

            * https://www.openssl.org/docs/man1.1.0/apps/ciphers.html

        :param client_ca: Client CA file for client-based auth.

            .. note: You can prepend ! (exclamation mark) to make client certificate
                authentication mandatory.

        :param wildcard: Allow regular expressions in ``name`` (used for wildcard certificates).

        """
        command = 'sni'

        if wildcard:
            command += '-regexp'

        args = [item for item in (cert, key, ciphers, client_ca) if item is not None]

        self._set(command, f"{name} {','.join(args)}")

        return self._section

    def set_sni_dir_params(self, dir, ciphers=None):
        """Enable checking for cert/key/client_ca file in the specified directory
        and create a sni/ssl context on demand.

        Expected filenames:
            * <sni-name>.crt
            * <sni-name>.key
            * <sni-name>.ca - this file is optional

        * http://uwsgi.readthedocs.io/en/latest/SNI.html#massive-sni-hosting

        :param str dir:

        :param str ciphers: Ciphers [alias] string.

            Example:
                * DEFAULT
                * HIGH
                * DHE, EDH

            * https://www.openssl.org/docs/man1.1.0/apps/ciphers.html

        """
        self._set('sni-dir', dir)
        self._set('sni-dir-ciphers', ciphers)

        return self._section



from ..base import OptionsGroup


class Applications(OptionsGroup):
    """Applications.

    """

    def set_basic_params(
            self,
            *,
            exit_if_none: bool = None,
            max_per_worker: int = None,
            single_interpreter: bool = None,
            no_default: bool = None,
            manage_script_name: bool = None
    ):
        """

        :param exit_if_none: Exit if no app can be loaded.

        :param max_per_worker: Set the maximum number of per-worker applications.

        :param single_interpreter: Do not use multiple interpreters (where available).
            Some of the supported languages (such as Python) have the concept of "multiple interpreters".
            By default every app is loaded in a new python interpreter (that means a pretty-well isolated
            namespace for each app). If you want all of the app to be loaded in the same python vm,
            use the this option.

        :param no_default: Do not automatically fallback to default app. By default, the first loaded app
            is mounted as the "default one". That app will be served when no mountpoint matches.

        :param manage_script_name: You can to instruct uWSGI to map specific apps in the so called "mountpoint"
            and rewrite SCRIPT_NAME and PATH_INFO automatically. See .mount().
            The WSGI standard dictates that SCRIPT_NAME is the variable used to select a specific application.

        """
        self._set('need-app', exit_if_none, cast=bool)
        self._set('max-apps', max_per_worker)
        self._set('single-interpreter', single_interpreter, cast=bool)
        self._set('no-default-app', no_default, cast=bool)
        self._set('manage-script-name', manage_script_name, cast=bool)

        return self._section

    def mount(self, mountpoint: str, app: str, *, into_worker: bool = False):
        """Load application under mountpoint.

        Example:
            * .mount('', 'app0.py') -- Root URL part
            * .mount('/app1', 'app1.py') -- URL part
            * .mount('/pinax/here', '/var/www/pinax/deploy/pinax.wsgi')
            * .mount('the_app3', 'app3.py')  -- Variable value: application alias (can be set by ``UWSGI_APPID``)
            * .mount('example.com', 'app2.py')  -- Variable value: Hostname (variable set in nginx)

        * http://uwsgi-docs.readthedocs.io/en/latest/Nginx.html#hosting-multiple-apps-in-the-same-process-aka-managing-script-name-and-path-info

        :param mountpoint: URL part, or variable value.

            .. note:: In case of URL part you may also want to set ``manage_script_name`` basic param to ``True``.

            .. warning:: In case of URL part a trailing slash may case problems in some cases
                (e.g. with Django based projects).

        :param app: App module/file.

        :param into_worker: Load application under mountpoint
            in the specified worker or after workers spawn.

        """
        # todo check worker mount -- uwsgi_init_worker_mount_app() expects worker://
        self._set('worker-mount' if into_worker else 'mount', f'{mountpoint}={app}', multi=True)

        return self._section

    def switch_into_lazy_mode(self, *, affect_master: bool = None):
        """Load apps in workers instead of master.

        This option may have memory usage implications
        as Copy-on-Write semantics can not be used.

        .. note:: Consider using ``touch_chain_reload`` option in ``workers`` basic params
            for lazy apps reloading.

        :param affect_master: If **True** only workers will be
          reloaded by uWSGI's reload signals; the master will remain alive.

          .. warning:: uWSGI configuration changes are not picked up on reload by the master.


        """
        self._set('lazy' if affect_master else 'lazy-apps', True, cast=bool)

        return self._section



import importlib.util
import inspect
import os
from importlib import import_module
from typing import Optional

from uwsgiconf.presets.nice import PythonSection
from uwsgiconf.settings import CONFIGS_MODULE_ATTR
from uwsgiconf.utils import ConfModule, UwsgiRunner

if False:  # pragma: nocover
    from uwsgiconf.base import Section  # noqa


def find_project_dir() -> str:
    """Runs up the stack to find the location of manage.py
    which will be considered a project base path.

    """
    frame = inspect.currentframe()

    while True:
        frame = frame.f_back
        filename = frame.f_globals['__file__']

        if os.path.basename(filename).startswith('manage.py'):  # support py, pyc, etc.
            break

    return os.path.dirname(filename)


def get_project_name(project_dir: str) -> str:
    """Return project name from project directory.

    :param project_dir:

    """
    return os.path.basename(project_dir)


class SectionMutator:
    """Configuration file section mutator."""

    def __init__(self, section: 'Section', dir_base: str, project_name: str, options: dict):
        from django.conf import settings

        self.section = section
        self.dir_base = dir_base
        self.project_name = project_name
        self.settings = settings
        self.options = options

    @property
    def runtime_dir(self) -> str:
        """Project runtime directory."""
        return self.section.replace_placeholders('{project_runtime_dir}')

    def get_pid_filepath(self) -> str:
        """Return pidfile path for the given project."""
        return os.path.join(self.runtime_dir, 'uwsgi.pid')

    def get_fifo_filepath(self) -> str:
        """Return master FIFO path for the given project."""
        return os.path.join(self.runtime_dir, 'uwsgi.fifo')

    @classmethod
    def spawn(cls, options: dict = None, dir_base: str = None) -> 'SectionMutator':
        """Alternative constructor. Creates a mutator and returns section object.

        :param options:
        :param dir_base:

        """
        options_all = {
            'compile': True,
            'embedded': False,

            'contribute_static': False,
            'contribute_runtimes': False,
            'contribute_errpages': False,
        }
        options_all.update(options or {})

        dir_base = os.path.abspath(dir_base or find_project_dir())

        name_module = ConfModule.default_name
        name_project = get_project_name(dir_base)
        path_conf = os.path.join(dir_base, name_module)
        embedded = options_all['embedded']

        # Read an existing config for further modification of first section.
        section = cls._get_section_existing(
            path_conf, name_module, name_project,
            embedded=embedded)

        if not section:
            # Create section on-fly.
            section = cls._get_section_new(dir_base)

        mutator = cls(
            section=section,
            dir_base=dir_base,
            project_name=name_project,
            options=options_all)

        mutator.mutate(embedded=embedded)

        return mutator

    @classmethod
    def _get_section_existing(
            self,
            path_conf: str,
            name_module: str,
            name_project: str,
            embedded: bool = False
    ) -> Optional['Section']:
        """Loads config section from existing configuration file (aka uwsgicfg.py)

        :param path_conf: Path containing configuration module.

        :param name_module: Configuration module name.

        :param name_project: Project (package) name.

        :param embedded: Flag. Do not try to load module file from file system manually,
            but try to import the module.

        """
        def load():
            module_fake_name = f'{name_project}.{os.path.splitext(name_module)[0]}'

            spec = importlib.util.spec_from_file_location(module_fake_name, path_conf)
            module = importlib.util.module_from_spec(spec)
            spec.loader.exec_module(module)

            return module

        if embedded:
            try:
                module = import_module(f"{name_project}.{name_module.rstrip('.py')}")

            except ModuleNotFoundError:
                return None

        else:
            if os.path.exists(path_conf):
                module = load()
            else:
                return None

        config = getattr(module, CONFIGS_MODULE_ATTR)[0]
        section = config.sections[0]

        return section

    @classmethod
    def _get_section_new(cls, dir_base: str) -> PythonSection:
        """Creates a new section with default settings.

        :param dir_base:

        """
        from django.conf import settings

        wsgi_app = settings.WSGI_APPLICATION
        name_package, name_module, name_func = wsgi_app.split('.')

        section = PythonSection.bootstrap(
            'http://127.0.0.1:8000',
            wsgi_module=f'{name_package}.{name_module}',
        )

        if os.path.exists(dir_base):
            section.main_process.change_dir(dir_base)

        return section

    def contribute_static(self):
        """Contributes static and media file serving settings to an existing section."""
        settings = self.settings
        statics = self.section.statics

        static_tuples = (
            (settings.STATIC_URL, settings.STATIC_ROOT),
            (settings.MEDIA_URL, settings.MEDIA_ROOT),
        )
        for url, path in static_tuples:
            path and statics.register_static_map(url, path)

        if self.options['compile']:
            return

        from django.core.management import call_command
        call_command('collectstatic', clear=True, interactive=False)

    def contribute_error_pages(self):
        """Contributes generic static error massage pages to an existing section."""

        static_dir = self.settings.STATIC_ROOT

        if not static_dir:
            # Source static directory is not configured. Use temporary.
            import tempfile
            static_dir = os.path.join(tempfile.gettempdir(), self.project_name)
            self.settings.STATIC_ROOT = static_dir

        self.section.routing.set_error_pages(
            common_prefix=os.path.join(static_dir, 'uwsgify'))

    def contribute_runtime_dir(self):
        section = self.section

        if not section.get_runtime_dir(default=False):
            # If runtime directory is not set by user, let's try use system default.
            section.set_runtime_dir(section.get_runtime_dir())

            if not self.options['compile']:
                os.makedirs(self.runtime_dir, 0o755, True)

    def mutate(self, embedded: bool = False):
        """Mutates current section."""
        section = self.section
        project_name = self.project_name

        section.project_name = project_name

        main = section.main_process

        if embedded:

            # The following should prevent possible segfaults in uwsgi_set_processname()'s memsets
            # while embedded.
            main.set_naming_params(autonaming=False)

            # Applications registry is ready by now,
            # we import base uwsgiinit in master process to bootstrap.
            section.python.import_module(
                'uwsgiconf.contrib.django.uwsgify.uwsgiinit',
                shared=True,
            )

        else:
            main.set_naming_params(prefix=f'[{project_name}] ')

        section.print_out(
            f"Embedded mode: {'yes' if embedded else 'no'}",
            format_options='blue')

        # todo maybe autoreload in debug

        apps = section.applications
        apps.set_basic_params(
            manage_script_name=True,
        )

        options = self.options

        if options['contribute_runtimes']:

            self.contribute_runtime_dir()

            main.set_pid_file(
                self.get_pid_filepath(),
                before_priv_drop=False,  # For vacuum to cleanup properly.
                safe=True,
            )

            section.master_process.set_basic_params(
                fifo_file=self.get_fifo_filepath(),
            )

        if options['contribute_static']:
            self.contribute_static()

        if options['contribute_errpages']:
            self.contribute_error_pages()


def run_uwsgi(config_section: 'Section', compile_only: bool = False, embedded: bool = False):
    """Runs uWSGI using the given section configuration.

    :param config_section:

    :param compile_only: Do not run, only compile and output configuration file for run.

    :param embedded: Do not create temporary config files and try to use resource files for configuration.

    """
    config = config_section.as_configuration()

    if compile_only:
        config.print_ini()
        return

    runner = UwsgiRunner()
    runner.spawn(
        config=config,
        replace=True,
        embedded=embedded,
    )



from typing import Union

from .signals import _automate_signal, Signal
from .. import uwsgi
from ..typehints import Strint


def register_file_monitor(filename: str, *, target: Union[Strint, Signal] = None):
    """Maps a specific file/directory modification event to a signal.

    :param filename: File or a directory to watch for its modification.

    :param target: Existing signal to raise
        or Signal Target to register signal implicitly.

        Available targets:

            * ``workers``  - run the signal handler on all the workers
            * ``workerN`` - run the signal handler only on worker N
            * ``worker``/``worker0`` - run the signal handler on the first available worker
            * ``active-workers`` - run the signal handlers on all the active [non-cheaped] workers

            * ``mules`` - run the signal handler on all of the mules
            * ``muleN`` - run the signal handler on mule N
            * ``mule``/``mule0`` - run the signal handler on the first available mule

            * ``spooler`` - run the signal on the first available spooler
            * ``farmN/farm_XXX``  - run the signal handler in the mule farm N or named XXX

    :raises ValueError: If unable to register monitor.

    """
    return _automate_signal(target, func=lambda sig: uwsgi.add_file_monitor(int(sig), filename))


class Metric:
    """User metric related stuff.

    .. note:: One needs to register user metric beforehand.
        E.g.:: ``section.monitoring.register_metric(section.monitoring.metric_types.absolute('mymetric'))``

    """
    def __init__(self, name: str):
        """
        :param name: Metric name.

        """
        self.name = name

    @property
    def value(self) -> int:
        """Current metric value."""
        return uwsgi.metric_get(self.name)

    def set(self, value: int, *, mode: str = None) -> bool:
        """Sets metric value.

        :param value: New value.

        :param mode: Update mode.

            * None - Unconditional update.
            * max - Sets metric value if it is greater that the current one.
            * min - Sets metric value if it is less that the current one.

        """
        if mode == 'max':
            func = uwsgi.metric_set_max

        elif mode == 'min':
            func = uwsgi.metric_set_min

        else:
            func = uwsgi.metric_set

        return func(self.name, value)

    def incr(self, delta: int = 1) -> bool:
        """Increments the specified metric key value by the specified value.

        :param delta:

        """
        return uwsgi.metric_inc(self.name, delta)

    def decr(self, delta: int = 1) -> bool:
        """Decrements the specified metric key value by the specified value.

        :param delta:

        """
        return uwsgi.metric_dec(self.name, delta)

    def mul(self, value: int = 1) -> bool:
        """Multiplies the specified metric key value by the specified value.

        :param value:

        """
        return uwsgi.metric_mul(self.name, value)

    def div(self, value: int = 1) -> bool:
        """Divides the specified metric key value by the specified value.

        :param value:

        """
        return uwsgi.metric_div(self.name, value)



from uwsgiconf.runtime.signals import *
from uwsgiconf.runtime.scheduling import register_timer


def test_signals():

    available = get_available_num()
    latest = get_last_received()

    sig = Signal()

    @sig.register_handler()
    def signalled():
        pass

    sig.send()
    sig.wait()

    @sig.register_handler()
    def my(sign):
        pass

    register_timer(3, target=sig)



from os import path

from django.core.management.base import BaseCommand

from uwsgiconf.sysinit import get_config, TYPE_SYSTEMD
from uwsgiconf.utils import Finder
from ...toolbox import SectionMutator


class Command(BaseCommand):

    help = 'Generates configuration files for Systemd, Upstart, etc.'

    def add_arguments(self, parser):  # pragma: nocover

        super().add_arguments(parser)

        parser.add_argument(
            '--systype', dest='systype',
            help='System type alias to make configuration for. E.g.: systemd, upstart.',
        )
        parser.add_argument(
            '--nostatic', action='store_true', dest='nostatic',
            help='Do not serve static and media files.',
        )
        parser.add_argument(
            '--noruntimes', action='store_true', dest='noruntimes',
            help='Do not automatically use a runtime directory to store pid and fifo files.',
        )
        parser.add_argument(
            '--noerrpages', action='store_true', dest='noerrpages',
            help='Do not to configure custom error pages (403, 404, 500).',
        )

    def handle(self, *args, **options):
        systype = options['systype'] or TYPE_SYSTEMD

        mutator = SectionMutator.spawn()
        command = 'manage.py uwsgi_run'

        for opt in ('nostatic', 'noruntimes', 'noerrpages'):
            if options.get(opt, False):
                command = command + f' --{opt}'

        config = get_config(
            systype,
            conf=mutator.section,
            conf_path=path.join(mutator.dir_base, command),
            runner=Finder.python(),
        )

        print(config)



import os

import pytest

from uwsgiconf.utils import UwsgiRunner, parse_command_plugins_output, ConfModule, get_uwsgi_stub_attrs_diff, \
    filter_locals
from uwsgiconf.exceptions import UwsgiconfException


SAMPLE_OUT_PLUGINS_MANY = b'''

*** uWSGI loaded generic plugins ***
gevent
syslog
stats_pusher_socket

*** uWSGI loaded request plugins ***
0: python
173: rpc
--- end of plugins list ---

*** Starting uWSGI 2.0.15 (64bit) on [Fri Jul 28 20:05:53 2017] ***
compiled with version: 6.3.0 20170406 on 28 July 2017 20:04:29
'''

SAMPLE_OUT_PLUGINS_EMPTY = b'''

*** uWSGI loaded generic plugins ***

*** uWSGI loaded request plugins ***
--- end of plugins list ---

*** Starting uWSGI 2.0.14-debian (64bit) on [Fri Jul 28 20:09:00 2017] ***
compiled with version: 6.3.0 20170221 on 27 February 2017 15:11:38
'''


def test_filter_locals():

    fake_locals = {'a': 1, 'b': 2, 'c': 3}

    assert filter_locals(fake_locals, drop=['a', 'c']) == {'b': 2}
    assert filter_locals(fake_locals, include=['a', 'b']) == {'a': 1, 'b': 2}
    assert filter_locals(fake_locals, include=['a', 'b'], drop=['b']) == {'a': 1}


def test_parser():
    plugins = parse_command_plugins_output(SAMPLE_OUT_PLUGINS_MANY.decode())

    assert len(plugins.generic) == 3
    assert len(plugins.request) == 2
    assert 'rpc' in plugins.request

    plugins = parse_command_plugins_output(SAMPLE_OUT_PLUGINS_EMPTY.decode())
    assert len(plugins.generic) == 0
    assert len(plugins.request) == 0


def test_runner(mock_popen):

    mock_popen(lambda: (SAMPLE_OUT_PLUGINS_MANY, ''))

    runner = UwsgiRunner()
    plugins = runner.get_plugins()
    assert len(plugins.generic) == 3
    assert len(plugins.request) == 2


def test_conf_module_compile():
    fpath = os.path.join(os.path.dirname(__file__), 'confs', 'dummy.py')

    # invalid objects
    module = ConfModule(fpath)
    assert module.configurations
    assert len(module.configurations) == 2

    out = []
    for conf in module.configurations:
        out.append(conf.format())

    out = '\n'.join(out)

    assert '[uwsgi]' in out
    assert '[conf1_2]' in out
    assert 'D=E' in out


def test_conf_module_run(monkeypatch):
    executed = []
    monkeypatch.setattr(os, 'spawnvp', lambda *args: executed.append(True))

    module = ConfModule(os.path.join(os.path.dirname(__file__), 'confs', 'dummy.py'))
    module.spawn_uwsgi()

    assert len(executed) == 2
    assert all(executed)

    executed = []
    monkeypatch.setattr(os, 'execvp', lambda *args: executed.append(True))

    module = ConfModule(os.path.join(os.path.dirname(__file__), 'confs', 'dummyone.py'))
    module.spawn_uwsgi()

    assert len(executed) == 1
    assert all(executed)


def test_get_uwsgi_stub_attrs_diff():

    with pytest.raises(UwsgiconfException):
        get_uwsgi_stub_attrs_diff()  # uwsgi is unavailable

    import sys

    sys.modules['uwsgi'] = type('DummyModule', (object,), {'dummy_uwsgi': True})

    from_candidate, from_stub = get_uwsgi_stub_attrs_diff()

    assert from_candidate == ['dummy_uwsgi']
    assert from_stub

    del sys.modules['uwsgi']


def test_maintenanceapp_import():
    from uwsgiconf.maintenance import app_maintenance

    responses = []
    def respond(*args):
        responses.append(args)

    assert b'Service is temporarily unavailable' in app_maintenance({'REQUEST_URI': '/'}, respond)
    assert app_maintenance({'REQUEST_URI': '/ping'}, respond) == b'success'

    assert responses == [
        ('503 Service Unavailable', [('Content-Type', 'text/html')]),
        ('200 OK', [('Content-Type', 'text/html')])
    ]



from django.core.management.base import BaseCommand

from ...toolbox import run_uwsgi, SectionMutator


class Command(BaseCommand):

    help = 'Runs uWSGI to serve your project'

    def add_arguments(self, parser):  # pragma: nocover

        super().add_arguments(parser)

        parser.add_argument(
            '--nostatic', action='store_false', dest='contribute_static',
            help='Do not serve static and media files.',
        )
        parser.add_argument(
            '--noruntimes', action='store_false', dest='contribute_runtimes',
            help='Do not automatically use a runtime directory to store pid and fifo files.',
        )
        parser.add_argument(
            '--noerrpages', action='store_false', dest='contribute_errpages',
            help='Do not to configure custom error pages (403, 404, 500).',
        )
        parser.add_argument(
            '--compile', action='store_true', dest='compile',
            help='Do not run just print out compiled uWSGI .ini configuration.',
        )
        parser.add_argument(
            '--embedded', action='store_true', dest='embedded',
            help='Do not create temporary config files and try to use resource files for configuration',
        )

    def handle(self, *args, **options):
        mutator = SectionMutator.spawn(options=options)
        run_uwsgi(
            mutator.section,
            compile_only=options['compile'],
            embedded=options['embedded'],
        )



from ._base import FifoCommand, Fifo


class Command(FifoCommand):

    help = 'Allows managing of uWSGI log related stuff'

    def add_arguments(self, parser):  # pragma: nocover

        super().add_arguments(parser)

        parser.add_argument(
            '--reopen', action='store_true', dest='reopen',
            help='Reopen log file. Could be required after third party rotation.',
        )
        parser.add_argument(
            '--rotate', action='store_true', dest='rotate',
            help='Trigger built-in log rotation.',
        )

    def run_cmd(self, fifo: Fifo, options: dict):
        fifo.cmd_log(reopen=options['reopen'], rotate=options['rotate'])



from uwsgiconf.runtime.caching import *


def test_caching():

    cache = Cache('mine')

    assert cache.keys == []

    cache.clear()

    cache.set('my', '10')
    cache.incr('my')
    cache.decr('my')
    cache.mul('my')
    cache.div('my')
    cache.delete('my')

    assert cache.get('some', default='other') == 'other'
    assert cache.get('int', default=10, as_int=True) == 10

    assert 'some' not in cache

    def article_picker(key):
        if key == '1':
            return 'one'

        elif key == '2':
            return 'two'

        return None

    assert cache.get('1', setter=article_picker) == 'one'
    assert cache.get('2', setter=article_picker) == 'two'
    assert cache.get('44', default=50, setter=article_picker) == 50



from ..base import TemplatedValue


class Var(TemplatedValue):

    pass


class Func(Var):

    pass


class VarGeoip(Var):
    """Returns Geoip data.

    http://uwsgi.readthedocs.io/en/latest/GeoIP.html

    """
    tpl = '${geoip[%s]}'

    COUNTRY_CODE = 'country_code'
    COUNTRY_CODE3 = 'country_code3'
    COUNTRY_NAME = 'country_name'
    CONTINENT = 'continent'
    REGION = 'region'
    REGION_NAME = 'region_name'
    CITY = 'city'
    POSTAL_CODE = 'postal_code'
    LATITUDE = 'lat'
    LONGITUDE = 'lon'
    DMA = 'dma'
    AREA = 'area'

    vars_country = [
        COUNTRY_CODE,
        COUNTRY_CODE3,
        COUNTRY_NAME,
    ]
    """Keys avaiable for country database."""

    vars_city = [
        CONTINENT,
        COUNTRY_CODE,
        COUNTRY_CODE3,
        COUNTRY_NAME,
        REGION,
        REGION_NAME,
        CITY,
        POSTAL_CODE,
        LATITUDE,
        LONGITUDE,
        DMA,
        AREA,
    ]
    """Keys avaiable for city database."""


class VarRequest(Var):
    """Returns request variable. Examples: PATH_INFO, SCRIPT_NAME, REQUEST_METHOD."""
    tpl = '${%s}'


class VarMetric(Var):
    """Returns metric (see ``monitoring``) variable."""
    tpl = '${metric[%s]}'


class VarCookie(Var):
    """Returns cookie variable"""
    tpl = '${cookie[%s]}'


class VarQuery(Var):
    """Returns query string variable."""
    tpl = '${qs[%s]}'


class VarUwsgi(Var):
    """Returns internal uWSGI information.

     Supported variables:
        * wid
        * pid
        * uuid
        * status

    """
    tpl = '${uwsgi[%s]}'


class VarTime(Var):
    """Returns time/date in various forms.

    Supported variables:
        * unix

    """
    tpl = '${time[%s]}'


class VarHttptime(Var):
    """Returns http date adding the numeric argument (if specified)
    to the current time (use empty arg for current server time).

    """
    tpl = '${httptime[%s]}'


class FuncMime(Func):
    """Returns mime type of a variable."""
    tpl = '${mime[%s]}'


class FuncMath(Func):
    """Perform a math operation. Example: CONTENT_LENGTH+1

    Supported operations: + - * /

    .. warning:: Requires matheval support.

    """
    tpl = '${math[%s]}'


class FuncBase64(Func):
    """Encodes the specified var in base64"""
    tpl = '${base64[%s]}'


class FuncHex(Func):
    """Encodes the specified var in hex."""
    tpl = '${hex[%s]}'


class FuncUpper(Func):
    """Uppercase the specified var."""
    tpl = '${upper[%s]}'


class FuncLower(Func):
    """Lowercase the specified var."""
    tpl = '${lower[%s]}'



from uwsgiconf.runtime.monitoring import *


def test_metric():

    m = Metric('mycounter')

    value = m.value

    m.set(10)
    m.set(20, mode='min')
    m.set(5, mode='max')

    m.incr()
    m.decr()
    m.mul()
    m.div()


def test_file_monitor():

    @register_file_monitor('/here/there.file')
    def handle_file_modification(sig_num):
        pass



from uwsgiconf.config import Section


def test_http(assert_lines):

    routers = Section.routing.routers

    router = routers.http(
        on='127.0.0.1:3111'
    ).set_basic_params(
        keepalive=20
    ).set_connections_params(
        harakiri=20
    ).set_manage_params(
        websockets=True
    ).set_owner_params('idle')

    assert_lines([
        'http = 127.0.0.1:3111',
        'http-keepalive = 20',
        'http-harakiri = 20',
        'http-websockets = true',
        'http-uid = idle',
        ],
        Section().routing.use_router(router)
    )

    router = routers.https(
        on='127.0.0.1:3112',
        cert='/here/mysert.cer',
        key='mykey',
        session_context='somectx')

    assert_lines([
        'https2 = cert=/here/mysert.cer,key=mykey,addr=127.0.0.1:3112',
        'http-session-context = somectx',
        ],
        Section().routing.use_router(router)
    )


def test_ssl(assert_lines):

    routers = Section.routing.routers

    router = routers.ssl(
        on='127.0.0.1:3112',
        cert='/here/mysert.cer',
        key='mykey',
    ).set_connections_params(
        retry_max=5
    )

    assert_lines([
        'sslrouter2 = cert=/here/mysert.cer,key=mykey,addr=127.0.0.1:3112',
        'sslrouter-max-retries = 5',
        ],
        Section().routing.use_router(router)
    )


def test_raw(assert_lines):

    routers = Section.routing.routers

    router = routers.raw(
        on='127.0.0.1:3115'
    ).set_connections_params(
        retry_max=5
    )

    assert_lines([
        'rawrouter = 127.0.0.1:3115',
        'rawrouter-max-retries = 5',
        ],
        Section().routing.use_router(router)
    )


def test_fast(assert_lines):

    routers = Section.routing.routers

    router = routers.fast(
        on='127.0.0.1:3113',
        forward_to=['127.0.0.1:3123', '127.0.0.1:3124'],
    ).set_connections_params(
        defer=5
    ).set_resubscription_params(
        addresses='127.0.0.1:3114'
    ).set_postbuffering_params(
        size=100
    ).set_owner_params('idle')

    assert_lines([
        'fastrouter = 127.0.0.1:3113',
        'fastrouter-to = 127.0.0.1:3123',
        'fastrouter-to = 127.0.0.1:3124',
        'fastrouter-defer-connect-timeout = 5',
        'fastrouter-resubscribe = 127.0.0.1:3114',
        'fastrouter-post-buffering = 100',
        'fastrouter-uid = idle',
        ],
        Section().routing.use_router(router)
    )


def test_forkpty(assert_lines):

    routers = Section.routing.routers

    router = routers.forkpty(
        on='127.0.0.1:3118',
        undeferred=True
    ).set_basic_params(
        run_command='/bin/zsh'
    ).set_connections_params(
        timeout_socket=13
    ).set_window_params(cols=10, rows=15)

    assert_lines([
        'forkptyurouter = 127.0.0.1:3118',
        'forkptyrouter-timeout = 13',
        'forkptyrouter-cols = 10',
        'forkptyrouter-rows = 15',
        ],
        Section().routing.use_router(router)
    )


def test_tuntap(assert_lines):

    routers = Section.routing.routers

    router = routers.tuntap(
        on='/tmp/tuntap.socket',
        device='emperor0',
        stats_server='127.0.0.1:3030',
        gateway='127.0.0.1:3032',
    ).set_basic_params(
        use_credentials='some'
    ).add_firewall_rule(
        direction='out',
        action='allow',
        src='192.168.0.0/24',
        dst='192.168.0.1',
    ).register_route(
        src='192.168.0.1/24',
        dst='192.168.0.2',
        gateway='192.168.0.4',
    )

    assert_lines([
        'tuntap-use-credentials = some',
        'tuntap-router = emperor0 /tmp/tuntap.socket 127.0.0.1:3030 127.0.0.1:3032',
        'tuntap-router-firewall-out = allow 192.168.0.0/24 192.168.0.1',
        'tuntap-router-route = 192.168.0.1/24 192.168.0.2 192.168.0.4',
        ],
        Section().routing.use_router(router)
    )

    router = routers.tuntap().device_connect(
        device_name='uwsgi0',
        socket='/tmp/tuntap.socket',
    ). device_add_rule(
        direction='in',
        action='route',
        src='192.168.0.1',
        dst='192.168.0.2',
        target='10.20.30.40:5060',
    )

    assert_lines([
        'tuntap-device = uwsgi0 /tmp/tuntap.socket',
        'tuntap-device-rule = in 192.168.0.1 192.168.0.2 route 10.20.30.40:5060',
        ],
        Section().routing.use_router(router)
    )


def test_forwarders(assert_lines):

    routers = Section.routing.routers

    router = routers.fast
    assert_lines(
        'fastrouter-use-base = /tmp/sockets/',

        Section().routing.use_router(router(on='127.0.0.1:3110',
            forward_to=router.forwarders.path('/tmp/sockets/'))))

    assert_lines(
        'fastrouter-use-pattern = /tmp/sockets/%s/uwsgi.sock',

        Section().routing.use_router(router(on='127.0.0.1:3110',
            forward_to=router.forwarders.path('/tmp/sockets/%s/uwsgi.sock'))))

    assert_lines(
        'fastrouter-use-code-string = 0:some.py:funcname',

        Section().routing.use_router(router(on='127.0.0.1:3110',
            forward_to=router.forwarders.code('some.py', 'funcname'))))

    assert_lines(
        'fastrouter-use-cache = mine',

        Section().routing.use_router(router(on='127.0.0.1:3110',
            forward_to=router.forwarders.cache('mine'))))

    assert_lines(
        'fastrouter-use-socket = /there/here.sock',

        Section().routing.use_router(router(on='127.0.0.1:3110',
            forward_to=router.forwarders.socket('/there/here.sock'))))

    assert_lines(
        'fastrouter-subscription-server = 192.168.0.100:7000',

        Section().routing.use_router(router(on='127.0.0.1:3110',
            forward_to=router.forwarders.subscription_server('192.168.0.100:7000'))))



from operator import attrgetter
from typing import List

from .alarm_types import AlarmType, AlarmCommand, AlarmSignal, AlarmCurl, AlarmLog, AlarmMule, AlarmXmpp
from ..base import OptionsGroup
from ..utils import listify

TypeAlarmExt = List[AlarmType]


class Alarms(OptionsGroup):
    """Alarms.

    This subsystem allows the developer/sysadmin to "announce"
    special conditions of an app via various channels.

    * http://uwsgi-docs.readthedocs.io/en/latest/AlarmSubsystem.html

    """
    class alarm_types:
        """Alarm types available for ``.register_alarm()``."""

        command = AlarmCommand
        curl = AlarmCurl
        log = AlarmLog
        mule = AlarmMule
        signal = AlarmSignal
        xmpp = AlarmXmpp

    def __init__(self, *args, **kwargs):
        self._alarms = []
        super().__init__(*args, **kwargs)

    def set_basic_params(self, *, msg_size: int = None, cheap: bool = None, anti_loop_timeout: int = None):
        """
        :param msg_size: Set the max size of an alarm message in bytes. Default: 8192.

        :param cheap: Use main alarm thread rather than create dedicated
            threads for curl-based alarms

        :param anti_loop_timeout: Tune the anti-loop alarm system. Default: 3 seconds.

        """
        self._set('alarm-msg-size', msg_size)
        self._set('alarm-cheap', cheap, cast=bool)
        self._set('alarm-freq', anti_loop_timeout)

        return self._section

    def print_alarms(self):
        """Print out enabled alarms."""

        self._set('alarm-list', True, cast=bool)

        return self._section

    def register_alarm(self, alarm: TypeAlarmExt):
        """Register (create) an alarm.

        :param alarm: Alarm.

        """
        for alarm in listify(alarm):
            if alarm not in self._alarms:
                self._set('alarm', alarm, multi=True)
                self._alarms.append(alarm)

        return self._section

    def alarm_on_log(self, alarm: TypeAlarmExt, matcher: str, *, skip: bool = False):
        """Raise (or skip) the specified alarm when a log line matches the specified regexp.

        :param alarm: Alarm.

        :param matcher: Regular expression to match log line.

        :param skip:

        """
        self.register_alarm(alarm)

        value = f"{','.join(map(attrgetter('alias'), listify(alarm)))} {matcher}"

        self._set('not-alarm-log' if skip else 'alarm-log', value)

        return self._section

    def alarm_on_fd_ready(self, alarm: TypeAlarmExt, *, fd: str, message: str, byte_count: int = None):
        """Triggers the alarm when the specified file descriptor is ready for read.

        This is really useful for integration with the Linux eventfd() facility.
        Pretty low-level and the basis of most of the alarm plugins.

        * http://uwsgi-docs.readthedocs.io/en/latest/Changelog-1.9.7.html#alarm-fd

        :param alarm: Alarm.

        :param fd: File descriptor.

        :param message: Message to send.

        :param byte_count: Files to read. Default: 1 byte.

            .. note:: For ``eventfd`` set 8.

        """
        self.register_alarm(alarm)

        value = fd

        if byte_count:
            value += f':{byte_count}'

        value += f' {message}'

        for alarm_ in listify(alarm):
            self._set('alarm-fd', f'{alarm_.alias} {value}', multi=True)

        return self._section

    def alarm_on_queue_full(self, alarm: TypeAlarmExt):
        """Raise the specified alarm when the socket backlog queue is full.

        :param alarm: Alarm.

        """
        self.register_alarm(alarm)

        for alarm_ in listify(alarm):
            self._set('alarm-backlog', alarm_.alias, multi=True)

        return self._section

    def alarm_on_segfault(self, alarm: TypeAlarmExt):
        """Raise the specified alarm when the segmentation fault handler is executed.

        Sends a backtrace.

        :param alarm: Alarm.

        """
        self.register_alarm(alarm)

        for alarm_ in listify(alarm):
            self._set('alarm-segfault', alarm_.alias, multi=True)

        return self._section



from uwsgiconf.config import configure_uwsgi


def get_configurations():
    from uwsgiconf.presets.nice import PythonSection
    section = PythonSection(name='testdummy')
    return section


configure_uwsgi(get_configurations)



import pytest


from uwsgiconf.options.routing_modifiers import ModifierCache
from uwsgiconf.exceptions import ConfigurationError


def test_routing_modifiers():

    with pytest.raises(ConfigurationError):
        ModifierCache(44)



import os

from uwsgiconf.presets.nice import PythonSection
from uwsgiconf.sysinit import get_config


def test_sysinit_systemd():

    section = PythonSection()
    config = get_config('systemd', conf=section, conf_path='/some/path')

    assert ('-o %s -g %s' % (os.getuid(), os.getgid())) in config
    assert 'run /some/path' in config
    assert 'Description=some uWSGI Service' in config

    section.main_process.set_owner_params(uid=9999, gid=9999)
    config = get_config('systemd', conf=section, conf_path='/some/path')
    assert '-o 9999 -g 9999' in config


def test_sysinit_upstart():

    section = PythonSection()

    config = get_config('upstart', conf=section, conf_path='/other/path')

    assert 'run /other/path' in config
    assert 'description "other uWSGI Service"' in config





from datetime import timedelta
from uwsgiconf.runtime.spooler import Spooler, ResultProcessed, ResultRescheduled, ResultSkipped, spooler_task_types


class FakeUwsgi(object):

    def __init__(self, send_to_spooler=None):
        self.send_func = send_to_spooler

    @property
    def opt(self):
        return {b'spooler': ['/home/here/spooler1', '/home/here/spooler2']}

    def spooler_pids(self):
        return [1, 2]

    def set_spooler_frequency(self, n):
        return n

    def spooler_jobs(self):
        return ['job1', 'job2']

    def spooler_get_task(self, filepath):
        return {'fpath': filepath}

    def send_to_spooler(self, message):

        raw = {
            b'spooler_task_name': b'sometaskname',
        }

        msg = message.get(b'msg')
        if msg:
            raw[b'msg'] = msg

        body = message.get(b'body')
        if body:
            raw[b'body'] = body

        return Spooler._process_message_raw(raw)


def test_spooler_basic(monkeypatch):

    assert spooler_task_types

    monkeypatch.setattr('uwsgiconf.runtime.spooler.uwsgi', FakeUwsgi())

    assert len(Spooler.get_spoolers()) == 2
    assert Spooler.get_pids() == [1, 2]
    assert Spooler.set_period(10) == 10
    assert Spooler.get_tasks() == ['job1', 'job2']
    assert Spooler.read_task_file('filepath') == {'fpath': 'filepath'}

    spooler = Spooler.get_by_basename('spooler2')
    assert isinstance(spooler, Spooler)
    assert str(spooler) == '/home/here/spooler2'

    assert Spooler.send_message_raw('a' * (64 * 1024), spooler=spooler) == ResultProcessed.code_uwsgi


def test_spooler_tasks(monkeypatch):

    faked = FakeUwsgi()
    monkeypatch.setattr('uwsgiconf.runtime.spooler.uwsgi', faked)

    spooler2 = Spooler.get_by_basename('spooler2')

    @Spooler.task(priority=3, postpone=timedelta(seconds=20))
    def spooler1_func(arg1, arg2=10):
        return arg1 + arg2

    @spooler2.task()
    def spooler2_func():
        return 'done'

    @spooler2.task()
    def spooler2_func_fail():
        raise Exception('damn')

    @spooler2.task()
    def spooler2_func_skip():
        return None

    assert spooler1_func(100, 20) == ResultProcessed.code_uwsgi
    assert spooler2_func() == ResultProcessed.code_uwsgi
    assert spooler2_func_skip() == ResultSkipped.code_uwsgi
    assert spooler2_func_fail() == ResultRescheduled.code_uwsgi



import os

from django.core.management.base import BaseCommand, CommandError

from uwsgiconf.utils import Fifo
from ...toolbox import SectionMutator


class FifoCommand(BaseCommand):
    """Base for uWSGI control management commands using master FIFO."""

    def run_cmd(self, fifo: Fifo, options: dict):
        """Must return FIFO command.

        :param fifo:
        :param options:

        """
        raise NotImplementedError

    def handle(self, *args, **options):
        mutator = SectionMutator.spawn()
        filepath = mutator.get_fifo_filepath()

        fifo = Fifo(filepath)

        if os.path.exists(filepath):

            self.run_cmd(fifo, options)

        else:
            raise CommandError(
                'Unable to find uWSGI FIFO file '
                f'for "{mutator.section.project_name}" project in {filepath}'
            )



from uwsgiconf.config import Section


def test_spooler_basics(assert_lines):

    assert_lines([
        'touch-spoolers-reload = /here/a',
    ], Section().spooler.set_basic_params(touch_reload='/here/a'))

    assert_lines([
        'spooler = /var/run/mine/one',
        'spooler = /var/run/mine/two',

    ], Section(
        runtime_dir='/var/run/', project_name='mine'
    ).spooler.add(work_dir=['{project_runtime_dir}/one', '{project_runtime_dir}/two']))

    assert_lines([
        'spooler-external = home/two',
    ], Section().spooler.add('home/two', external=True))

    assert_lines([
        'spooler = /base/here/alias1',
        'spooler = /base/here/alias2',
    ], Section().spooler.set_basic_params(base_dir='/base/here').spooler.add(['alias1', 'alias2']))



from ..base import ParametrizedValue
from ..utils import KeyValue, filter_locals
from .monitoring_metric_types import Metric


class Collector(ParametrizedValue):

    name_separator = ','
    name_separator_strip = True
    args_joiner = ','

    def __init__(self, *args, **kwargs):
        args = list(args)

        # Adding children for metric to use.
        children = kwargs.pop('children', None)

        if children:
            children_ = []

            for child in children:
                if isinstance(child, Metric):
                    child = child.name
                children_.append(child)

            args.append(f"children={';'.join(children_)}")

        super().__init__(*args)


class CollectorPointer(Collector):
    """The value is collected from memory pointer."""

    name = 'ptr'


class CollectorFile(Collector):
    """The value is collected from a file."""

    name = 'file'

    def __init__(self, fpath, *, get_slot=None):
        """
        :param str fpath: File path.

        :param int get_slot: Get value from the given slot number.
            Slots: the content is split (using \\n, \\t, spaces, \\r and zero as separator)
            and the item (the returned array is zero-based) used as the return value.

        """
        value = KeyValue(locals(), aliases={'fpath': 'arg1', 'get_slot': 'arg1n'})

        super().__init__(value)


class CollectorFunction(Collector):
    """The value is computed calling a specific C function every time.

    .. note::
        * The argument it takes is a ``uwsgi_metric`` pointer.
          You generally do not need to parse the metric, so just casting to void will avoid headaches.

        * The function must returns an ``int64_t`` value.

    """

    name = 'func'

    def __init__(self, func):
        """
        :param str func: Function to call.
        """
        value = KeyValue(locals(), aliases={'func': 'arg1'})

        super().__init__(value)


class CollectorSum(Collector):
    """The value is the sum of other metrics."""

    name = 'sum'

    def __init__(self, what):
        super().__init__(children=what)


class CollectorAvg(Collector):
    """The value is the algebraic average of the children.

    .. note:: Since 1.9.20

    """
    name = 'avg'

    def __init__(self, what):
        super().__init__(children=what)


class CollectorAccumulator(Collector):
    """Always add the sum of children to the final value.

    Example:

        * Round 1: child1 = 22, child2 = 17 -> metric_value = 39
        * Round 2: child1 = 26, child2 = 30 -> metric_value += 56

    """
    name = 'accumulator'

    def __init__(self, what):
        super().__init__(children=what)


class CollectorAdder(Collector):
    """Add the specified argument (arg1n) to the sum of children.

    """
    name = 'adder'

    def __init__(self, what, value):
        """
        :param int value: Value to add (multiply if it is CollectorMultiplier).

        """
        value = KeyValue(filter_locals(locals(), drop=['what']), aliases={'value': 'arg1n'})

        super().__init__(value, children=what)


class CollectorMultiplier(CollectorAdder):
    """Multiply the sum of children by the specified argument.

    Example:

        * child1 = 22, child2 = 17, arg1n = 3 -> metric_value = (22+17)*3

    """
    name = 'multiplier'



from os import path

from ..base import OptionsGroup
from ..typehints import Strlist
from ..utils import listify


class Spooler(OptionsGroup):
    """Spooler.

    .. note:: Supported on: Perl, Python, Ruby.

    .. note:: Be sure the ``spooler`` plugin is loaded in your instance,
        but generally it is built in by default.

    The Spooler is a queue manager built into uWSGI that works like a printing/mail system.
    You can enqueue massive sending of emails, image processing, video encoding, etc.
    and let the spooler do the hard work in background while your users
    get their requests served by normal workers.

    http://uwsgi-docs.readthedocs.io/en/latest/Spooler.html

    """

    def __init__(self, *args, **kwargs):
        self._base_dir = ''
        super().__init__(*args, **kwargs)

    def set_basic_params(
            self,
            *,
            touch_reload: Strlist = None,
            quiet: bool = None,
            process_count: int = None,
            max_tasks: int = None,
            order_tasks: int = None,
            harakiri: int = None,
            change_dir: str = None,
            poll_interval: int = None,
            signal_as_task: bool = None,
            cheap: bool = None,
            base_dir: str = None
    ):
        """

        :param touch_reload: reload spoolers if the specified file is modified/touched

        :param quiet: Do not log spooler related messages.

        :param process_count: Set the number of processes for spoolers.

        :param max_tasks: Set the maximum number of tasks to run before recycling
            a spooler (to help alleviate memory leaks).

        :param order_tasks: Try to order the execution of spooler tasks (uses scandir instead of readdir).

        :param harakiri: Set harakiri timeout for spooler tasks.

        :param change_dir: chdir() to specified directory before each spooler task.

        :param poll_interval: Spooler poll frequency in seconds. Default: 30.

        :param signal_as_task: Treat signal events as tasks in spooler.
            To be used with ``spooler-max-tasks``. If enabled spooler will treat signal
            events as task. Run signal handler will also increase the spooler task count.

        :param cheap: Use spooler cheap mode.

        :param base_dir: Base directory to prepend to `work_dir` argument of `.add()`.

        """
        self._set('touch-spoolers-reload', touch_reload, multi=True)
        self._set('spooler-quiet', quiet, cast=bool)
        self._set('spooler-processes', process_count)
        self._set('spooler-max-tasks', max_tasks)
        self._set('spooler-ordered', order_tasks, cast=bool)
        self._set('spooler-harakiri', harakiri)
        self._set('spooler-chdir', change_dir)
        self._set('spooler-frequency', poll_interval)
        self._set('spooler-signal-as-task', signal_as_task, cast=bool)
        self._set('spooler-cheap', cheap, cast=bool)

        if base_dir is not None:
            self._base_dir = base_dir

        return self._section

    def add(self, work_dir: Strlist, *, external: bool = False):
        """Run a spooler on the specified directory.

        :param work_dir: Spooler working directory path or it's name if
            `base_dir` argument of `spooler.set_basic_params()` is set.

            .. note:: Placeholders can be used to build paths, e.g.: {project_runtime_dir}/spool/
              See ``Section.project_name`` and ``Section.runtime_dir``.

        :param external: map spoolers requests to a spooler directory managed by an external instance

        """
        command = 'spooler'

        base_dir = self._base_dir
        if base_dir is not None:
            work_dir = [path.join(base_dir, work_dir_) for work_dir_ in listify(work_dir)]

        if external:
            command += '-external'

        self._set(command, self._section.replace_placeholders(work_dir), multi=True)

        return self._section




def app_maintenance(env, start_response):
    """This will answer requests while in maintenance mode."""
    content_type = [('Content-Type', 'text/html')]

    if env.get('REQUEST_URI', '') == '/ping':
        # This may satisfy readiness probes, used in certain cloud services.
        start_response('200 OK', content_type)
        return b'success'

    start_response('503 Service Unavailable', content_type)
    return b'Service is temporarily unavailable due to a maintenance.<br>Please come back later.'



from uwsgiconf.runtime.rpc import *


def test_rpc():

    @register_rpc()
    def expose_me():
        pass

    make_rpc_call('expose_me')
    make_rpc_call('remotefunc', remote='@faraway')

    get_rpc_list()



from datetime import datetime
from functools import partial, wraps
from typing import Union, Callable

from .signals import _automate_signal, Signal
from .. import uwsgi
from ..exceptions import RuntimeConfigurationError
from ..typehints import Strint

TypeTarget = Union[Strint, Signal]
TypeRegResult = Union[Callable, bool]


def register_timer(period: int, *, target: TypeTarget = None) -> TypeRegResult:
    """Add timer.

    Can be used as a decorator:

        .. code-block:: python

            @register_timer(3)
            def repeat():
                do()

    :param int period: The interval (seconds) at which to raise the signal.

    :param target: Existing signal to raise or Signal Target to register signal implicitly.

        Available targets:

            * ``workers``  - run the signal handler on all the workers
            * ``workerN`` - run the signal handler only on worker N
            * ``worker``/``worker0`` - run the signal handler on the first available worker
            * ``active-workers`` - run the signal handlers on all the active [non-cheaped] workers

            * ``mules`` - run the signal handler on all of the mules
            * ``muleN`` - run the signal handler on mule N
            * ``mule``/``mule0`` - run the signal handler on the first available mule

            * ``spooler`` - run the signal on the first available spooler
            * ``farmN/farm_XXX``  - run the signal handler in the mule farm N or named XXX

    :raises ValueError: If unable to add timer.

    """
    return _automate_signal(target, func=lambda sig: uwsgi.add_timer(int(sig), period))


def register_timer_rb(period: int, *, repeat: int = None, target: TypeTarget = None) -> TypeRegResult:
    """Add a red-black timer (based on black-red tree).

        .. code-block:: python

            @register_timer_rb(3)
            def repeat():
                do()

    :param period: The interval (seconds) at which the signal is raised.

    :param repeat: How many times to send signal. Will stop after ther number is reached.
        Default: None - infinitely.

    :param target: Existing signal to raise or Signal Target to register signal implicitly.

        Available targets:

            * ``workers``  - run the signal handler on all the workers
            * ``workerN`` - run the signal handler only on worker N
            * ``worker``/``worker0`` - run the signal handler on the first available worker
            * ``active-workers`` - run the signal handlers on all the active [non-cheaped] workers

            * ``mules`` - run the signal handler on all of the mules
            * ``muleN`` - run the signal handler on mule N
            * ``mule``/``mule0`` - run the signal handler on the first available mule

            * ``spooler`` - run the signal on the first available spooler
            * ``farmN/farm_XXX``  - run the signal handler in the mule farm N or named XXX

    :raises ValueError: If unable to add timer.

    """
    return _automate_signal(target, func=lambda sig: uwsgi.add_rb_timer(int(sig), period, repeat or 0))


def register_timer_ms(period: int, *, target: TypeTarget = None) -> TypeRegResult:
    """Add a millisecond resolution timer.

        .. code-block:: python

            @register_timer_ms(300)
            def repeat():
                do()

    :param period: The interval (milliseconds) at which the signal is raised.

    :param target: Existing signal to raise or Signal Target to register signal implicitly.

        Available targets:

            * ``workers``  - run the signal handler on all the workers
            * ``workerN`` - run the signal handler only on worker N
            * ``worker``/``worker0`` - run the signal handler on the first available worker
            * ``active-workers`` - run the signal handlers on all the active [non-cheaped] workers

            * ``mules`` - run the signal handler on all of the mules
            * ``muleN`` - run the signal handler on mule N
            * ``mule``/``mule0`` - run the signal handler on the first available mule

            * ``spooler`` - run the signal on the first available spooler
            * ``farmN/farm_XXX``  - run the signal handler in the mule farm N or named XXX

    :raises ValueError: If unable to add timer.

    """
    return _automate_signal(target, func=lambda sig: uwsgi.add_ms_timer(int(sig), period))


def register_cron(
        *,
        weekday: Strint = None,
        month: Strint = None,
        day: Strint = None,
        hour: Strint = None,
        minute: Strint = None,
        target: TypeTarget = None
) -> TypeRegResult:
    """Adds cron. The interface to the uWSGI signal cron facility.

        .. code-block:: python

            @register_cron(hour=-3)  # Every 3 hours.
            def repeat():
                do()

    .. note:: Arguments work similarly to a standard crontab,
        but instead of "*", use -1,
        and instead of "/2", "/3", etc. use -2 and -3, etc.

    .. note:: Periods - rules like hour='10-18/2' (from 10 till 18 every 2 hours) - are allowed,
        but they are emulated by uwsgiconf. Use strings to define periods.

        Keep in mind, that your actual function will be wrapped into another one, which will check
        whether it is time to call your function.

    :param weekday: Day of a the week number. Defaults to `each`.
        0 - Sunday  1 - Monday  2 - Tuesday  3 - Wednesday
        4 - Thursday  5 - Friday  6 - Saturday

    :param month: Month number 1-12. Defaults to `each`.

    :param day: Day of the month number 1-31. Defaults to `each`.

    :param hour: Hour 0-23. Defaults to `each`.

    :param minute: Minute 0-59. Defaults to `each`.

    :param target: Existing signal to raise or Signal Target to register signal implicitly.

        Available targets:

            * ``workers``  - run the signal handler on all the workers
            * ``workerN`` - run the signal handler only on worker N
            * ``worker``/``worker0`` - run the signal handler on the first available worker
            * ``active-workers`` - run the signal handlers on all the active [non-cheaped] workers

            * ``mules`` - run the signal handler on all of the mules
            * ``muleN`` - run the signal handler on mule N
            * ``mule``/``mule0`` - run the signal handler on the first available mule

            * ``spooler`` - run the signal on the first available spooler
            * ``farmN/farm_XXX``  - run the signal handler in the mule farm N or named XXX

    :raises ValueError: If unable to add cron rule.

    """
    task_args_initial = {name: val for name, val in locals().items() if val is not None and name != 'target'}
    task_args_casted = {}

    def skip_task(check_funcs):
        now = datetime.now()
        allright = all((func(now) for func in check_funcs))
        return not allright

    def check_date(now, attr, target_range):
        attr = getattr(now, attr)

        if callable(attr):  # E.g. weekday.
            attr = attr()

        return attr in target_range

    check_date_funcs = []

    for name, val in task_args_initial.items():

        # uWSGI won't accept strings, so we emulate ranges here.
        if isinstance(val, str):

            # Rules like 10-18/2 (from 10 till 18 every 2 hours).
            val, _, step = val.partition('/')
            step = int(step) if step else 1
            start, _, end = val.partition('-')

            if not (start and end):
                raise RuntimeConfigurationError(
                    'String cron rule without a range is not supported. Use integer notation.')

            start = int(start)
            end = int(end)

            now_attr_name = name

            period_range = set(range(start, end+1, step))

            if name == 'weekday':
                # Special case for weekday indexes: swap uWSGI Sunday 0 for ISO Sunday 7.
                now_attr_name = 'isoweekday'

                if 0 in period_range:
                    period_range.discard(0)
                    period_range.add(7)

            # Gather date checking functions in one place.
            check_date_funcs.append(partial(check_date, attr=now_attr_name, target_range=period_range))

            # Use minimal duration (-1).
            val = None

        task_args_casted[name] = val

    if not check_date_funcs:
        # No special handling of periods, quit early.
        args = [(-1 if arg is None else arg) for arg in (minute, hour, day, month, weekday)]
        return _automate_signal(target, func=lambda sig: uwsgi.add_cron(int(sig), *args))

    skip_task = partial(skip_task, check_date_funcs)

    def decor(func_action):
        """Decorator wrapping."""

        @wraps(func_action)
        def func_action_wrapper(*args, **kwargs):
            """Action function wrapper to handle periods in rules."""

            if skip_task():
                # todo Maybe allow user defined value for this return.
                return None

            return func_action(*args, **kwargs)

        args = []
        for arg_name in ['minute', 'hour', 'day', 'month', 'weekday']:
            arg = task_args_casted.get(arg_name, None)
            args.append(-1 if arg is None else arg)

        return _automate_signal(target, func=lambda sig: uwsgi.add_cron(int(sig), *args))(func_action_wrapper)

    return decor



from datetime import datetime, timedelta
from pathlib import Path

from django import forms
from django.contrib import messages
from django.template.defaultfilters import filesizeformat
from django.utils.translation import gettext_lazy as _

from .base import OnePageAdmin


class SummaryAdmin(OnePageAdmin):

    def contribute_onepage_context(self, request, context):
        from uwsgiconf.runtime.platform import uwsgi
        from uwsgiconf.runtime.logging import get_current_log_size
        from uwsgiconf.runtime.rpc import get_rpc_list
        from uwsgiconf.runtime.signals import registry_signals
        from uwsgiconf.runtime.spooler import Spooler
        from uwsgiconf.runtime.mules import Farm

        def get_func_name(func) -> str:
            """Returns a distinctive name for a given function."""
            module_path = func.__module__

            if module_path.startswith('uwsgi_file'):
                module_path = module_path.replace('uwsgi_file__', 'uwsgi://', 1).replace('_', '/')

            return f'{module_path}.{func.__name__}'

        def get_signals_info(signals):
            info = []

            for signal in signals:
                info.append(f'{signal.num} - {signal.target}: {get_func_name(signal.func)}')

            return info

        time_started = datetime.fromtimestamp(uwsgi.started_on)
        rss, vsz = uwsgi.memory
        config = uwsgi.config

        info_basic = {
            _('Version'): uwsgi.get_version(),
            _('Hostname'): uwsgi.hostname,
            _('Serving since'): time_started,
            _('Serving for'): datetime.now() - time_started,
            _('Clock'): uwsgi.clock,
            _('Master PID'): uwsgi.master_pid,
            _('Memory (RSS, VSZ)'): '\n'.join((filesizeformat(rss), filesizeformat(vsz))),
            _('Buffer size'): uwsgi.buffer_size,
            _('Cores'): uwsgi.cores_count,
            _('Workers'): uwsgi.workers_count,
            _('Mules'): config.get('mules', 0),
            _('Farms'): '\n'.join(map(str, Farm.get_farms())),
            _('Threads support'): '+' if uwsgi.threads_enabled else '-',
            _('Current worker'): uwsgi.worker_id,
            _('Requests by worker'): uwsgi.request.id,
            _('Requests total'): uwsgi.request.total_count,
            _('Socket queue size'): uwsgi.get_listen_queue(),
            _('Log size'): get_current_log_size(),
            _('RPC'): '\n'.join(get_rpc_list()),
            _('Post fork hooks'): '\n'.join(map(get_func_name, uwsgi.postfork_hooks.funcs)),
            _('Signals'): '\n'.join(get_signals_info(registry_signals)),
            _('Spoolers'): '\n'.join(map(str, Spooler.get_spoolers())),
        }

        context.update({
            'panels': {
                '': {
                    'rows': dict((key, [val]) for key, val in info_basic.items()),
                }
            },
        })


class ConfigurationAdmin(OnePageAdmin):

    def contribute_onepage_context(self, request, context):
        from uwsgiconf.runtime.platform import uwsgi

        context.update({
            'panels': {
                '': {'rows': dict((key, [val]) for key, val in uwsgi.config.items())},
            },
        })


class WorkersAdmin(OnePageAdmin):

    def contribute_onepage_context(self, request, context):
        from uwsgiconf.runtime.platform import uwsgi

        fromts = datetime.fromtimestamp

        info_worker_map = {
            'id': (_('ID'), None),
            'pid': (_('PID'), None),
            'status': (_('Status'), None),
            'running_time': (_('Running for'), lambda val: timedelta(microseconds=val)),
            'last_spawn': (_('Spawned at'), lambda val: fromts(val)),
            'respawn_count': (_('Respawns'), None),
            'requests': (_('Requests'), None),
            'delta_requests': (_('Delta requests'), None),  # Used alongside with MAX_REQUESTS
            'exceptions': (_('Exceptions'), None),
            'signals': (_('Signals'), None),
            'rss': (_('RSS'), lambda val: filesizeformat(val)),
            'vsz': (_('VSZ'), lambda val: filesizeformat(val)),
            'tx': (_('Transmitted'), lambda val: filesizeformat(val)),
            'avg_rt': (_('Avg. response'), lambda val: timedelta(microseconds=val)),
            'apps': (None, lambda val: iter_items(val, info_app_map)),
        }

        info_app_map = {
            'id': (_('ID'), None),
            'startup_time': (_('Serving since'), None),
            'interpreter': (_('Interpreter'), None),
            'modifier1': (_('Modifier 1'), None),
            'mountpoint': (_('Mountpoint'), None),
            'callable': (_('Callable'), None),
            'chdir': (_('Directory'), None),
            'requests': (_('Requests'), None),
            'exceptions': (_('Exceptions'), None),
        }

        panels = {}
        info_workers = {}
        panels[''] = {'rows': info_workers}

        info_apps = {}

        def iter_items(info, mapping):

            unknown = set()

            for idx, info_item in enumerate(info):
                for keyname, (name, func) in mapping.items():

                    value = info_item.get(keyname, unknown)
                    if value is unknown:
                        continue

                    if func is not None:
                        value = func(value)

                    yield idx, keyname, name, value

        for idx_worker, keyname_worker, name_worker, value_worker in iter_items(uwsgi.workers_info, info_worker_map):

            if keyname_worker == 'apps':
                # Get info about applications served by worker,
                for idx_app, keyname_app, name_app, value_app in value_worker:
                    app_key = f'%s {idx_worker + 1}. %s {idx_app}' % (_('Worker'), _('Application'))
                    info_apps.setdefault(app_key, {})[name_app] = [value_app]

            else:
                info_workers.setdefault(name_worker, []).append(value_worker)

        # Add panel for every app on every worker.
        for title, info in info_apps.items():
            panels[title] = {'rows': info}

        context.update({'panels': panels})


class MaintenanceForm(forms.Form):

    confirm_word = 'maintenance'

    confirm = forms.CharField(
        label=_('Confirmation'),
        help_text=_('Enter the word "%(word)s".') % {'word': confirm_word})

    def clean_confirm(self):

        value = self.cleaned_data['confirm']
        confirm_word = self.confirm_word

        if value != confirm_word:
            raise forms.ValidationError(
                _('You need to enter "%(word)s" to proceed.') % {'word': confirm_word})

        return value


class MaintenanceAdmin(OnePageAdmin):

    def contribute_onepage_context(self, request, context):

        from uwsgiconf.settings import get_maintenance_path
        maintenance = get_maintenance_path()

        if not maintenance:
            self.message_user(
                request,
                _('Maintenance mode is not supported in this setup.'),
                level=messages.WARNING
            )
            return

        form_kwargs = {}
        is_submitted = request.method == 'POST'

        if is_submitted:
            form_kwargs['data'] = request.POST

        form = MaintenanceForm(**form_kwargs)

        if is_submitted and form.is_valid():

            try:
                Path(maintenance).touch()
                self.message_user(request, _('Maintenance mode is scheduled.'))

            except Exception as e:
                self.message_user(
                    request,
                    _('Unable to schedule maintenance: %(error)s.') % {'error': f'{e}'})

        else:

            self.message_user(
                request,
                _('Remember: maintenance mode can only be turned off from the filesystem. '
                  'If the process is scheduled, your application, including this administration interface, '
                  'will only become available after filesystem manipulation and uWSGI restart.'),
                level=messages.WARNING
            )

            context.update({
                'panels': {'': {'rows': {'': [form]}}},
                'show_save': True,
            })



from typing import Tuple

from ..base import ParametrizedValue
from ..utils import listify


if False:  # pragma: nocover
    from .routing_modifiers import Modifier  # noqa


class Socket(ParametrizedValue):

    opt_key = ''
    args_joiner = ','

    def __init__(self, address, *, bound_workers=None, modifier=None):
        """
        :param str|SocketShared address: Address ([host]:port or socket file) to bind socket to.

        :param  str|int|list bound_workers: Map socket to specific workers.
            As you can bind a uWSGI instance to multiple sockets, you can use this option to map
            specific workers to specific sockets to implement a sort of in-process Quality of Service scheme.
            If you host multiple apps in the same uWSGI instance, you can easily dedicate resources to each of them.

        :param Modifier modifier: Socket routing modifier.

        """
        self.address = address
        self.bound_workers = listify(bound_workers or [])

        self._make_section_like()

        if modifier:
            self._set(f'{self.name}-modifier1', modifier.code)

            submod = modifier.submod

            if submod:
                self._set(f'{self.name}-modifier2', modifier.submod)

        super().__init__()

    def __str__(self):
        if self.address not in self.args:
            self.args.insert(0, self.address)

        result = super().__str__()

        self.args.pop(0)

        return result


class SocketDefault(Socket):
    """Bind using default protocol. See ``default_socket_type`` option."""

    name = 'socket'


class SocketHttp(Socket):
    """Bind to the specified socket using HTTP"""

    name = 'http-socket'

    def __init__(self, address, *, http11=False, bound_workers=None, modifier=None):
        """
        :param str|SocketShared address: Address ([host]:port or socket file) to bind socket to.

        :param bool http11: Keep-Alive support. If set the server will try to maintain
            the connection opened if a bunch of rules are respected.

            This is not a smart http 1.1 parser (to avoid parsing the whole response)
            but assumes the developer is generating the right headers.

            This has been added to support RTSP protocol for video streaming.

        :param  str|int|list bound_workers: Map socket to specific workers.
            As you can bind a uWSGI instance to multiple sockets, you can use this option to map
            specific workers to specific sockets to implement a sort of in-process Quality of Service scheme.
            If you host multiple apps in the same uWSGI instance, you can easily dedicate resources to each of them.

        :param Modifier modifier: Socket routing modifier.

        """
        if http11:
            self.name = 'http11-socket'

        super().__init__(address, bound_workers=bound_workers, modifier=modifier)


class SocketHttps(Socket):
    """Bind to the specified socket using HTTPS"""

    name = 'https-socket'

    def __init__(self, address, *, cert, key, ciphers=None, client_ca=None, bound_workers=None, modifier=None):
        """
        :param str|SocketShared address: Address ([host]:port or socket file) to bind socket to.

        :param str cert: Certificate file.

        :param str key: Private key file.

        :param str ciphers: Ciphers [alias] string.

            Example:
                * DEFAULT
                * HIGH
                * DHE, EDH

            * https://www.openssl.org/docs/man1.1.0/apps/ciphers.html

        :param str client_ca: Client CA file for client-based auth.

            .. note: You can prepend ! (exclamation mark) to make client certificate
                authentication mandatory.

        :param  str|int|list bound_workers: Map socket to specific workers.
            As you can bind a uWSGI instance to multiple sockets, you can use this option to map
            specific workers to specific sockets to implement a sort of in-process Quality of Service scheme.
            If you host multiple apps in the same uWSGI instance, you can easily dedicate resources to each of them.

        :param Modifier modifier: Socket routing modifier.
        """
        super().__init__(address, bound_workers=bound_workers, modifier=modifier)
        args = [cert, key]

        if ciphers or client_ca:
            args.extend([ciphers or '', client_ca or ''])

        self.args.extend(args)

    @classmethod
    def get_certbot_paths(cls, domain: str) -> Tuple[str, str]:
        """Returns a tuple of paths for files (certificates_chain, private_key)
        from Certbot https://certbot.eff.org

        Those paths can be used to pass into Socket initializer.

        .. note:: If files not found empty strings are returned.

        :param domain: Domain name to get filepaths for.

        """
        from pathlib import Path

        certs_root = Path('/etc/letsencrypt/live/')
        certs_chain = certs_root / domain / 'fullchain.pem'
        certs_private = certs_root / domain / 'privkey.pem'

        if certs_chain.exists() and certs_private.exists():
            return str(certs_chain), str(certs_private)

        return '', ''


class SocketUwsgi(Socket):
    """uwSGI specific socket using ``uwsgi`` protocol."""

    name = 'uwsgi-socket'

    def __init__(self, address, *, persistent=False, bound_workers=None, modifier=None):
        """
        :param str|SocketShared address: Address ([host]:port or socket file) to bind socket to.

        :param bool persistent: Use persistent uwsgi protocol (puwsgi).

        :param  str|int|list bound_workers: Map socket to specific workers.
            As you can bind a uWSGI instance to multiple sockets, you can use this option to map
            specific workers to specific sockets to implement a sort of in-process Quality of Service scheme.
            If you host multiple apps in the same uWSGI instance, you can easily dedicate resources to each of them.

        :param Modifier modifier: Socket routing modifier.

        """
        if persistent:
            self.name = 'puwsgi-socket'

        super().__init__(address, bound_workers=bound_workers, modifier=modifier)


class SocketUwsgis(SocketHttps):
    """uwSGI specific socket using ``uwsgi`` protocol over SSL."""

    name = 'suwsgi-socket'


class SocketUdp(Socket):
    """Run the udp server on the specified address.

    .. note:: Mainly useful for SNMP or shared UDP logging.

    """
    name = 'udp'


class SocketFastcgi(Socket):
    """Bind to the specified socket using FastCGI."""

    name = 'fastcgi-socket'

    def __init__(self, address, *, nph=False, bound_workers=None, modifier=None):
        """
        :param str|SocketShared address: Address ([host]:port or socket file) to bind socket to.

        :param bool nph: Use NPH mode ("no-parsed-header" - bypass the server completely by sending
            the complete HTTP header directly to the browser).

        :param  str|int|list bound_workers: Map socket to specific workers.
            As you can bind a uWSGI instance to multiple sockets, you can use this option to map
            specific workers to specific sockets to implement a sort of in-process Quality of Service scheme.
            If you host multiple apps in the same uWSGI instance, you can easily dedicate resources to each of them.

        :param Modifier modifier: Socket routing modifier.

        """
        if nph:
            self.name = 'fastcgi-nph-socket'

        super().__init__(address, bound_workers=bound_workers, modifier=modifier)


class SocketScgi(Socket):
    """Bind to the specified UNIX/TCP socket using SCGI protocol."""

    name = 'scgi-socket'

    def __init__(self, address, *, nph=False, bound_workers=None, modifier=None):
        """
        :param str|SocketShared address: Address ([host]:port or socket file) to bind socket to.

        :param bool nph: Use NPH mode ("no-parsed-header" - bypass the server completely by sending
            the complete HTTP header directly to the browser).

        :param  str|int|list bound_workers: Map socket to specific workers.
            As you can bind a uWSGI instance to multiple sockets, you can use this option to map
            specific workers to specific sockets to implement a sort of in-process Quality of Service scheme.
            If you host multiple apps in the same uWSGI instance, you can easily dedicate resources to each of them.

        :param Modifier modifier: Socket routing modifier.

        """
        if nph:
            self.name = 'scgi-nph-socket'

        super().__init__(address, bound_workers=bound_workers, modifier=modifier)


class SocketRaw(Socket):
    """Bind to the specified UNIX/TCP socket using RAW protocol.

    Raw mode allows you to directly parse the request in your application callable.
    Instead of getting a list of CGI vars/headers in your callable you only get
    the file descriptor soon after accept().

    You can then read()/write() to that file descriptor in full freedom.

    .. note:: Raw mode disables request logging.

    .. warning:: Use it as a low-level socket wrapper.

    """
    name = 'raw-socket'


class SocketShared(Socket):
    """Create a shared socket for advanced jailing or IPC purposes.

    Allows you to create a socket early in the server's startup
    and use it after privileges drop or jailing. This can be used
    to bind to privileged (<1024) ports.

    Shared sockets are a way to share sockets among various uWSGI components:
    you can use that to share a socket between the fastrouter and uWSGI instance.

    """
    name = 'shared-socket'

    def __init__(self, address, *, undeferred=False, bound_workers=None, modifier=None):
        """
        :param str address: Address ([host]:port or socket file) to bind socket to.

        :param bool undeferred: Use shared socket undeferred mode.

        :param  str|int|list bound_workers: Map socket to specific workers.
            As you can bind a uWSGI instance to multiple sockets, you can use this option to map
            specific workers to specific sockets to implement a sort of in-process Quality of Service scheme.
            If you host multiple apps in the same uWSGI instance, you can easily dedicate resources to each of them.

        :param Modifier modifier: Socket routing modifier.

        """
        if undeferred:
            self.name = 'undeferred-shared-socket'

        super().__init__(address, bound_workers=bound_workers, modifier=modifier)


class SocketZeromq(Socket):
    """Introduce zeromq pub/sub pair."""

    name = 'zeromq-socket'



from ..base import ParametrizedValue
from ..typehints import Strlist
from ..utils import listify, filter_locals, KeyValue


class AlarmType(ParametrizedValue):

    alias: str = ''

    def __init__(self, alias: str, *args, **kwargs):
        self.alias = alias or ''
        super().__init__(*args)


class AlarmCommand(AlarmType):
    """Run a shell command, passing info into its stdin."""

    name = 'cmd'

    def __init__(self, alias: str, *, command: str):
        super().__init__(alias, command)


class AlarmSignal(AlarmType):
    """Raise an uWSGI signal."""

    name = 'signal'

    def __init__(self, alias: str, *, sig: int):
        super().__init__(alias, sig)


class AlarmLog(AlarmType):
    """Print line into log."""

    name = 'log'

    def __init__(self, alias: str):
        super().__init__(alias)


class AlarmMule(AlarmType):
    """Send info to a mule waiting for messages."""

    name = 'mule'

    def __init__(self, alias: str, *, mule: int):
        super().__init__(alias, mule)


class AlarmCurl(AlarmType):
    """Send info to a cURL-able URL."""

    name = 'curl'
    plugin = 'alarm_curl'
    args_joiner = ';'

    def __init__(
            self,
            alias: str,
            url: str,
            *,
            method: str = None,
            ssl: bool = None,
            ssl_insecure: bool = None,
            auth_user: str = None,
            auth_pass: str = None,
            timeout: int = None,
            conn_timeout: int = None,
            mail_from: str = None,
            mail_to: str = None,
            subject: str = None
    ):
        opts = KeyValue(
            filter_locals(locals(), drop=['alias', 'url']),
            bool_keys=['ssl', 'ssl_insecure'],
            items_separator=self.args_joiner,
        )
        super().__init__(alias, url, opts)


class AlarmXmpp(AlarmType):
    """Send info via XMPP/jabber."""

    name = 'xmpp'
    plugin = 'alarm_xmpp'
    args_joiner = ';'

    def __init__(self, alias: str, *, jid: str, password: str, recipients: Strlist):
        super().__init__(alias, jid, password, ','.join(listify(recipients)))
